"""Support for August devices."""

from __future__ import annotations

import asyncio
import logging
from abc import abstractmethod
from collections.abc import Callable, Coroutine, Iterable, ValuesView
from contextlib import suppress
from datetime import datetime
from enum import Enum
from functools import partial
from itertools import chain
from typing import Any, ParamSpec, TypeVar

from aiohttp import ClientError, ClientResponseError, ClientSession

from .._compat import cached_property
from ..activity import Activity, ActivityTypes, Source
from ..backports.tasks import create_eager_task
from ..const import Brand
from ..doorbell import ContentTokenExpired, Doorbell, DoorbellDetail
from ..exceptions import AugustApiAIOHTTPError, YaleApiError
from ..lock import Lock, LockDetail, LockOperation
from ..pubnub_activity import activities_from_pubnub_message
from ..pubnub_async import AugustPubNub
from .activity import ActivityStream
from .const import MIN_TIME_BETWEEN_DETAIL_UPDATES
from .exceptions import CannotConnect, YaleXSError
from .gateway import Gateway
from .ratelimit import _RateLimitChecker
from .socketio import SocketIORunner
from .subscriber import SubscriberMixin

_LOGGER = logging.getLogger(__name__)

API_CACHED_ATTRS = {
    "door_state",
    "door_state_datetime",
    "lock_status",
    "lock_status_datetime",
}
YALEXS_BLE_DOMAIN = "yalexs_ble"

_R = TypeVar("_R")
_P = ParamSpec("_P")


class _PushUpdatesState(Enum):
    """Push updates connection state."""

    NOT_CONNECTED = "not_connected"
    CONNECTED = "connected"


def _save_live_attrs(lock_detail: DoorbellDetail | LockDetail) -> dict[str, Any]:
    """Store the attributes that the lock detail api may have an invalid cache for.

    Since we are connected to pubnub we may have more current data
    then the api so we want to restore the most current data after
    updating battery state etc.
    """
    return {attr: getattr(lock_detail, attr) for attr in API_CACHED_ATTRS}


def _restore_live_attrs(
    lock_detail: DoorbellDetail | LockDetail, attrs: dict[str, Any]
) -> None:
    """Restore the non-cache attributes after a cached update."""
    for attr, value in attrs.items():
        setattr(lock_detail, attr, value)


class YaleXSData(SubscriberMixin):
    """YaleXS Data coordinator object."""

    def __init__(
        self, gateway: Gateway, error_exception_class: Exception = YaleXSError
    ) -> None:
        """Init August data object."""
        super().__init__(MIN_TIME_BETWEEN_DETAIL_UPDATES)
        self._gateway = gateway
        self.activity_stream: ActivityStream = None
        self._api = gateway.api
        self._device_detail_by_id: dict[str, LockDetail | DoorbellDetail] = {}
        self._doorbells_by_id: dict[str, Doorbell] = {}
        self._locks_by_id: dict[str, Lock] = {}
        self._house_ids: set[str] = set()
        self._push_unsub: Callable[[], Coroutine[Any, Any, None]] | None = None
        self._initial_sync_task: asyncio.Task | None = None
        self._error_exception_class = error_exception_class
        self._shutdown: bool = False
        # Track last known state from WebSocket messages to avoid unnecessary updates
        self._last_push_state: dict[str, dict[str, str]] = {}

    @cached_property
    def brand(self) -> Brand:
        """Return the brand of the API."""
        return self._gateway.api.brand

    async def async_setup(self) -> None:
        """Async setup of august device data and activities."""
        token = await self._gateway.async_get_access_token()
        await _RateLimitChecker.check_rate_limit(token)
        await _RateLimitChecker.register_wakeup(token)

        # This used to be a gather but it was less reliable with august's recent api changes.
        locks: list[Lock] = await self._api.async_get_operable_locks(token) or []
        doorbells: list[Doorbell] = await self._api.async_get_doorbells(token) or []
        self._doorbells_by_id = {device.device_id: device for device in doorbells}
        self._locks_by_id = {device.device_id: device for device in locks}
        self._house_ids = {device.house_id for device in chain(locks, doorbells)}

        await self._async_refresh_device_detail_by_ids(
            [device.device_id for device in chain(locks, doorbells)]
        )

        # We remove all devices that we are missing
        # detail as we cannot determine if they are usable.
        # This also allows us to avoid checking for
        # detail being None all over the place
        self._remove_inoperative_locks()
        self._remove_inoperative_doorbells()

        # Fetch capabilities for locks (only for Yale brands)
        if self.brand in (Brand.YALE_GLOBAL, Brand.YALE_HOME):
            await self._async_fetch_lock_capabilities()

        await self.async_setup_activity_stream()

        if self._locks_by_id and self.brand is not Brand.YALE_GLOBAL:
            # Do not prevent setup as the sync can timeout
            # but it is not a fatal error as the lock
            # will recover automatically when it comes back online.
            self._initial_sync_task = create_eager_task(
                self._async_initial_sync(), name="august-initial-sync"
            )

    async def async_setup_activity_stream(self) -> None:
        """Set up the activity stream."""
        token = await self._gateway.async_get_access_token()
        user_data = await self._api.async_get_user(token)
        push: AugustPubNub | SocketIORunner
        if self.brand is Brand.YALE_GLOBAL:
            push = SocketIORunner(self._gateway)
            push_source = Source.WEBSOCKET
        else:
            push = AugustPubNub()
            push_source = Source.PUBNUB
            for device in self._device_detail_by_id.values():
                push.register_device(device)
        self.activity_stream = ActivityStream(
            self._api, self._gateway, self._house_ids, push
        )
        await self.activity_stream.async_setup()
        # Use partial to bind the source parameter
        push_callback = partial(self.async_push_message, source=push_source)
        push.subscribe(push_callback)
        self._push_unsub = await push.run(user_data["UserID"], self.brand)

    async def _async_fetch_lock_capabilities(self) -> None:
        """Fetch capabilities for all locks from the API."""
        token = await self._gateway.async_get_access_token()

        # Fetch capabilities for each lock in series to keep API load low
        for lock_id, lock_detail in self._device_detail_by_id.items():
            if lock_id not in self._locks_by_id or not isinstance(
                lock_detail, LockDetail
            ):
                continue

            # lock_id is the serial_number for locks
            try:
                capabilities = await self._api.async_get_lock_capabilities(
                    token, lock_id
                )
                lock_detail.set_capabilities(capabilities)
                _LOGGER.debug(
                    "Fetched capabilities for lock %s: unlatch=%s",
                    lock_detail.device_name,
                    capabilities.get("lock", {}).get("unlatch", False),
                )
            except YaleApiError as ex:
                # 409 Conflict means the API cannot determine device type from serial
                # 404 Not Found means device info not found
                # These can happen for older devices, just log debug and continue
                if ex.status in (404, 409):
                    _LOGGER.debug(
                        "Cannot fetch capabilities for lock %s (HTTP %s): %s",
                        lock_detail.device_name,
                        ex.status,
                        str(ex),
                    )
                else:
                    _LOGGER.warning(
                        "Failed to fetch capabilities for lock %s (HTTP %s): %s",
                        lock_detail.device_name,
                        ex.status,
                        str(ex),
                    )
            except (ClientError, TimeoutError) as ex:
                _LOGGER.warning(
                    "Failed to fetch capabilities for lock %s: %s",
                    lock_detail.device_name,
                    ex,
                )

    async def _async_initial_sync(self) -> None:
        """Attempt to request an initial sync."""
        # We don't care if this fails because we only want to wake
        # locks that are actually online anyways and they will be
        # awake when they come back online
        for result in await asyncio.gather(
            *[
                create_eager_task(
                    self._async_status_async(
                        device_id, bool(detail.bridge and detail.bridge.hyper_bridge)
                    )
                )
                for device_id, detail in self._device_detail_by_id.items()
                if device_id in self._locks_by_id
            ],
            return_exceptions=True,
        ):
            if isinstance(result, Exception) and not isinstance(
                result, (TimeoutError, ClientResponseError, CannotConnect)
            ):
                _LOGGER.warning(
                    "Unexpected exception during initial sync: %s",
                    result,
                    exc_info=result,
                )

    def async_push_message(
        self,
        device_id: str,
        date_time: datetime,
        message: dict[str, Any],
        source: Source | str = "unknown",
    ) -> None:
        """Process a push message."""
        try:
            self._async_handle_push_message(device_id, date_time, message, source)
        except Exception:
            _LOGGER.exception(
                "Error processing push message for device %s at %s: %s",
                device_id,
                date_time,
                message,
            )
            # If we have an error, we want to continue processing other messages
            return

    def _async_handle_push_message(
        self,
        device_id: str,
        date_time: datetime,
        message: dict[str, Any],
        source: Source | str,
    ) -> None:
        """Handle a push message."""
        _LOGGER.debug("async_push_message from %s: %s %s", source, device_id, message)
        device = self.get_device_detail(device_id)
        activities = activities_from_pubnub_message(device, date_time, message, source)

        # Check if this is a push message with unchanged state
        changed = not self._is_unchanged_push_state(
            device_id, message, source, activities
        )
        activity_stream = self.activity_stream
        _LOGGER.debug("async_push_message activities: %s for %s", activities, device_id)
        if activities and activity_stream.async_process_newer_device_activities(
            activities
        ):
            _LOGGER.debug(
                "async_push_message newer activities: %s for %s", device_id, activities
            )
            self.async_signal_device_id_update(device.device_id)
            if not changed:
                _LOGGER.debug(
                    "Skipping unchanged %s state for %s: status=%s, lockAction=%s, doorState=%s",
                    source,
                    device_id,
                    message.get("status"),
                    message.get("lockAction"),
                    message.get("doorState"),
                )
                return
            for activity in activities:
                # Don't trigger a house refresh if the activity is a status update
                # to avoid unnecessary API calls.
                if activity.is_status:
                    _LOGGER.debug(
                        "async_push_message activity: %s is status update",
                        activity,
                    )
                    continue
                _LOGGER.debug(
                    "async_push_message activity triggering refresh: %s for %s",
                    device_id,
                    activity,
                )
                activity_stream.async_schedule_house_id_refresh(device.house_id)
                break

    async def async_stop(self, *args: Any) -> None:
        """Stop the subscriptions."""
        self._shutdown = True
        if self.activity_stream:
            self.activity_stream.async_stop()
        if self._initial_sync_task:
            self._initial_sync_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._initial_sync_task
        if self._push_unsub:
            await self._push_unsub()

    @property
    def doorbells(self) -> ValuesView[Doorbell]:
        """Return a list of py-august Doorbell objects."""
        return self._doorbells_by_id.values()

    @property
    def locks(self) -> ValuesView[Lock]:
        """Return a list of py-august Lock objects."""
        return self._locks_by_id.values()

    def get_device_detail(self, device_id: str) -> DoorbellDetail | LockDetail:
        """Return the py-august LockDetail or DoorbellDetail object for a device."""
        return self._device_detail_by_id[device_id]

    async def _async_refresh(self) -> None:
        """Refresh data."""
        if self._shutdown:
            return
        await self._async_refresh_device_detail_by_ids(self._subscriptions.keys())

    async def _async_refresh_device_detail_by_ids(
        self, device_ids_list: Iterable[str]
    ) -> None:
        """Refresh each device in sequence.

        This used to be a gather but it was less reliable with august's
        recent api changes.

        The august api has been timing out for some devices so
        we want the ones that it isn't timing out for to keep working.
        """
        for device_id in device_ids_list:
            try:
                await self._async_refresh_device_detail_by_id(device_id)
            except TimeoutError:  # noqa: PERF203
                _LOGGER.warning(
                    "Timed out calling august api during refresh of device: %s",
                    device_id,
                )
            except (ClientResponseError, CannotConnect) as err:
                _LOGGER.warning(
                    "Error from august api during refresh of device: %s",
                    device_id,
                    exc_info=err,
                )

    async def refresh_camera_by_id(self, device_id: str) -> None:
        """Re-fetch doorbell/camera data from API."""
        await self._async_update_device_detail(
            self._doorbells_by_id[device_id],
            self._api.async_get_doorbell_detail,
        )

    @property
    def push_updates_connected(self) -> bool:
        """Return if the push updates are connected."""
        return (
            self.activity_stream is not None
            and self.activity_stream.push_updates_connected
        )

    async def _async_refresh_device_detail_by_id(self, device_id: str) -> None:
        if self._shutdown:
            return
        if device_id in self._locks_by_id:
            if self.activity_stream and self.push_updates_connected:
                saved_attrs = _save_live_attrs(self._device_detail_by_id[device_id])
            await self._async_update_device_detail(
                self._locks_by_id[device_id], self._api.async_get_lock_detail
            )
            if self.activity_stream and self.push_updates_connected:
                _restore_live_attrs(self._device_detail_by_id[device_id], saved_attrs)
            # keypads are always attached to locks
            if (
                device_id in self._device_detail_by_id
                and self._device_detail_by_id[device_id].keypad is not None
            ):
                keypad = self._device_detail_by_id[device_id].keypad
                self._device_detail_by_id[keypad.device_id] = keypad
        elif device_id in self._doorbells_by_id:
            await self._async_update_device_detail(
                self._doorbells_by_id[device_id],
                self._api.async_get_doorbell_detail,
            )
        _LOGGER.debug(
            "async_signal_device_id_update (from detail updates): %s", device_id
        )
        self.async_signal_device_id_update(device_id)

    async def _async_update_device_detail(
        self,
        device: Doorbell | Lock,
        api_call: Callable[
            [str, str], Coroutine[Any, Any, DoorbellDetail | LockDetail]
        ],
    ) -> None:
        device_id = device.device_id
        device_name = device.device_name
        _LOGGER.debug("Started retrieving detail for %s (%s)", device_name, device_id)

        try:
            detail = await api_call(
                await self._gateway.async_get_access_token(), device_id
            )
        except ClientError as ex:
            _LOGGER.error(
                "Request error trying to retrieve %s details for %s. %s",
                device_id,
                device_name,
                ex,
            )
        _LOGGER.debug("Completed retrieving detail for %s (%s)", device_name, device_id)
        # If the key changes after startup we need to trigger a
        # discovery to keep it up to date
        if isinstance(detail, LockDetail) and detail.offline_key:
            self.async_offline_key_discovered(detail)

        self._device_detail_by_id[device_id] = detail

    @abstractmethod
    def async_offline_key_discovered(self, detail: LockDetail) -> None:
        """Handle offline key discovery."""

    def get_device(self, device_id: str) -> Doorbell | Lock | None:
        """Get a device by id."""
        return self._locks_by_id.get(device_id) or self._doorbells_by_id.get(device_id)

    def _get_device_name(self, device_id: str) -> str | None:
        """Return doorbell or lock name as August has it stored."""
        if device := self.get_device(device_id):
            return device.device_name
        return None

    async def async_lock(self, device_id: str) -> list[ActivityTypes]:
        """Lock the device."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_lock_return_activities,
            await self._gateway.async_get_access_token(),
            device_id,
        )

    async def async_status_async(self, device_id: str, hyper_bridge: bool) -> str:
        """Request status of the device but do not wait for a response since it will come via pubnub."""
        token = await self._gateway.async_get_access_token()
        await _RateLimitChecker.check_rate_limit(token)
        result = await self._async_status_async(device_id, hyper_bridge)
        await _RateLimitChecker.register_wakeup(token)
        return result

    async def _async_status_async(self, device_id: str, hyper_bridge: bool) -> str:
        """Request status of the device but do not wait for a response since it will come via pubnub."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_status_async,
            await self._gateway.async_get_access_token(),
            device_id,
            hyper_bridge,
        )

    async def async_lock_async(self, device_id: str, hyper_bridge: bool) -> str:
        """Lock the device but do not wait for a response since it will come via pubnub."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_lock_async,
            await self._gateway.async_get_access_token(),
            device_id,
            hyper_bridge,
        )

    async def async_unlatch(self, device_id: str) -> list[ActivityTypes]:
        """Open/unlatch the device."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_unlatch_return_activities,
            await self._gateway.async_get_access_token(),
            device_id,
        )

    async def async_unlatch_async(self, device_id: str, hyper_bridge: bool) -> str:
        """Open/unlatch the device but do not wait for a response since it will come via pubnub."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_unlatch_async,
            await self._gateway.async_get_access_token(),
            device_id,
            hyper_bridge,
        )

    async def async_unlock(self, device_id: str) -> list[ActivityTypes]:
        """Unlock the device."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_unlock_return_activities,
            await self._gateway.async_get_access_token(),
            device_id,
        )

    async def async_unlock_async(self, device_id: str, hyper_bridge: bool) -> str:
        """Unlock the device but do not wait for a response since it will come via pubnub."""
        return await self._async_call_api_op_requires_bridge(
            device_id,
            self._api.async_unlock_async,
            await self._gateway.async_get_access_token(),
            device_id,
            hyper_bridge,
        )

    async def async_operate_lock(
        self,
        device_id: str,
        operation: LockOperation,
        push_updates_connected: bool = False,
        hyper_bridge: bool = True,
    ) -> list[ActivityTypes]:
        """Unified method to operate the lock (lock/unlock/open).

        Args:
            device_id: The device ID
            operation: The operation type (LockOperation enum)
            push_updates_connected: If True, use async operation that doesn't wait for response
            hyper_bridge: Whether to use hyper bridge for async operations

        Returns:
            Activities list when waiting for response, or empty list when using push updates

        Note:
            When unlatch is supported, unlock and open operations are swapped:
            - UNLOCK operation calls unlatch API
            - OPEN operation calls unlock API

        """
        # Base operation map - lock is always the same
        operation_map = {
            LockOperation.LOCK: {
                _PushUpdatesState.NOT_CONNECTED: self.async_lock,
                _PushUpdatesState.CONNECTED: self.async_lock_async,
            },
        }

        # Check if the device supports unlatching to determine unlock/open mapping
        detail = self.get_device_detail(device_id)
        if detail and detail.unlatch_supported:
            # When unlatch is supported, swap unlock and open operations
            operation_map[LockOperation.UNLOCK] = {
                _PushUpdatesState.NOT_CONNECTED: self.async_unlatch,  # Swapped!
                _PushUpdatesState.CONNECTED: self.async_unlatch_async,
            }
            operation_map[LockOperation.OPEN] = {
                _PushUpdatesState.NOT_CONNECTED: self.async_unlock,  # Swapped!
                _PushUpdatesState.CONNECTED: self.async_unlock_async,
            }
        else:
            # Normal mapping when unlatch is not supported
            operation_map[LockOperation.UNLOCK] = {
                _PushUpdatesState.NOT_CONNECTED: self.async_unlock,
                _PushUpdatesState.CONNECTED: self.async_unlock_async,
            }
            operation_map[LockOperation.OPEN] = {
                _PushUpdatesState.NOT_CONNECTED: self.async_unlatch,
                _PushUpdatesState.CONNECTED: self.async_unlatch_async,
            }

        if operation not in operation_map:
            raise ValueError(f"Invalid operation: {operation}")

        # Determine the push updates state
        state = (
            _PushUpdatesState.CONNECTED
            if push_updates_connected
            else _PushUpdatesState.NOT_CONNECTED
        )
        method = operation_map[operation][state]

        if push_updates_connected:
            await method(device_id, hyper_bridge)
            return []
        return await method(device_id)

    async def _async_call_api_op_requires_bridge(
        self,
        device_id: str,
        func: Callable[_P, Coroutine[Any, Any, _R]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _R:
        """Call an API that requires the bridge to be online and will change the device state."""
        try:
            ret = await func(*args, **kwargs)
        except AugustApiAIOHTTPError as err:
            device_name = self._get_device_name(device_id)
            if device_name is None:
                device_name = f"DeviceID: {device_id}"
            raise self._error_exception_class(f"{device_name}: {err}") from err

        return ret

    async def async_get_doorbell_image(
        self,
        device_id: str,
        aiohttp_session: ClientSession,
        timeout: float = 10.0,
    ) -> bytes:
        """Get the latest image from the doorbell."""
        doorbell = self.get_device_detail(device_id)
        try:
            return await doorbell.async_get_doorbell_image(aiohttp_session, timeout)
        except ContentTokenExpired:
            if self.brand not in (Brand.YALE_HOME, Brand.YALE_GLOBAL):
                raise
            _LOGGER.debug(
                "Error fetching camera image, updating content-token from api to retry"
            )
            await self.refresh_camera_by_id(device_id)
            doorbell = self.get_device_detail(device_id)
            return await doorbell.async_get_doorbell_image(aiohttp_session, timeout)

    def _remove_inoperative_doorbells(self) -> None:
        for doorbell in list(self.doorbells):
            device_id = doorbell.device_id
            if self._device_detail_by_id.get(device_id):
                continue
            _LOGGER.info(
                (
                    "The doorbell %s could not be setup because the system could not"
                    " fetch details about the doorbell"
                ),
                doorbell.device_name,
            )
            del self._doorbells_by_id[device_id]

    def _is_unchanged_push_state(
        self,
        device_id: str,
        message: dict[str, Any],
        source: Source | str,
        activities: list[Activity],
    ) -> bool:
        """Check if a push message represents unchanged state."""
        # Build state key based on source - different sources track state separately
        state_key = f"{device_id}:{source}"
        # Get last known state
        last_state = self._last_push_state.get(state_key)

        # Get relevant state fields based on source
        if source == Source.WEBSOCKET:
            # WebSocket uses lockAction and doorState
            if "lockAction" not in message and "doorState" not in message:
                return False
            current_state = {
                "lock": message.get("lockAction", ""),
                "door": message.get("doorState", ""),
            }
        else:  # PubNub
            # PubNub uses status and doorState
            if "status" not in message and "doorState" not in message:
                return False
            current_state = {
                "lock": message.get("status", ""),
                "door": message.get("doorState", ""),
            }
            # If all activities are status updates and we have a last state,
            # check if the state is unchanged but don't update tracking
            if last_state and all(activity.is_status for activity in activities):
                # Status update with changed state or no previous state - don't track it
                return last_state == current_state

        # If we have a previous state and it matches current state, it's unchanged
        if last_state and last_state == current_state:
            return True

        # Update the last known state (only for real actions, not status updates)
        self._last_push_state[state_key] = current_state
        return False

    def _remove_inoperative_locks(self) -> None:
        # Remove non-operative locks as there must
        # be a bridge (August Connect) for them to
        # be usable
        for lock in list(self.locks):
            device_id = lock.device_id
            lock_detail = self._device_detail_by_id.get(device_id)
            if lock_detail is None:
                _LOGGER.info(
                    (
                        "The lock %s could not be setup because the system could not"
                        " fetch details about the lock"
                    ),
                    lock.device_name,
                )
            elif lock_detail.bridge is None:
                _LOGGER.info(
                    (
                        "The lock %s could not be setup because it does not have a"
                        " bridge (Connect)"
                    ),
                    lock.device_name,
                )
                del self._device_detail_by_id[device_id]
            # Bridge may come back online later so we still add the device since we will
            # have a pubnub subscription to tell use when it recovers
            else:
                continue
            del self._locks_by_id[device_id]



from __future__ import annotations

import asyncio
from unittest.mock import AsyncMock, MagicMock

import pytest
from freezegun.api import FrozenDateTimeFactory

from yalexs.api_async import ApiAsync
from yalexs.manager.activity import (
    ACTIVITY_DEBOUNCE_COOLDOWN,
    INITIAL_LOCK_RESYNC_TIME,
    UPDATE_SOON,
    ActivityStream,
)
from yalexs.manager.gateway import Gateway

from ..common import fire_time_changed


@pytest.mark.asyncio
async def test_activity_stream_debounce(freezer: FrozenDateTimeFactory) -> None:
    """Test activity stream debounce."""

    api = MagicMock(auto_spec=ApiAsync)
    async_get_house_activities = AsyncMock()
    api.async_get_house_activities = async_get_house_activities
    august_gateway = MagicMock(auto_spec=Gateway)
    august_gateway.async_refresh_access_token_if_needed = AsyncMock()
    august_gateway.async_get_access_token = AsyncMock()
    push = MagicMock(connected=False)
    august_gateway.push = push

    activity = ActivityStream(api, august_gateway, {"myhouseid"}, push)
    await activity.async_setup()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    assert async_get_house_activities.call_count == 1
    freezer.tick(INITIAL_LOCK_RESYNC_TIME)
    fire_time_changed()
    assert async_get_house_activities.call_count == 1
    async_get_house_activities.reset_mock()
    assert "myhouseid" not in activity._schedule_updates

    activity.async_schedule_house_id_refresh("myhouseid")
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 0
    freezer.tick(UPDATE_SOON)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    assert "myhouseid" not in activity._schedule_updates
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    assert "myhouseid" not in activity._schedule_updates

    activity.async_schedule_house_id_refresh("myhouseid")
    await asyncio.sleep(0)
    assert activity._pending_updates["myhouseid"] == 2
    assert async_get_house_activities.call_count == 2
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 3
    assert activity._pending_updates["myhouseid"] == 1

    # If we get another update request, be sure we reset
    # but we do not poll right away and only do 2 polls
    activity.async_schedule_house_id_refresh("myhouseid")
    await asyncio.sleep(0)
    assert activity._pending_updates["myhouseid"] == 2
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert activity._pending_updates["myhouseid"] == 1
    assert async_get_house_activities.call_count == 4
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert activity._pending_updates["myhouseid"] == 0
    assert async_get_house_activities.call_count == 5
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert activity._pending_updates["myhouseid"] == 0
    assert async_get_house_activities.call_count == 5
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 5
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)

    # If we get another update request later, be sure we reset
    # and poll after 1s with 3 polls
    activity.async_schedule_house_id_refresh("myhouseid")
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 5
    assert activity._pending_updates["myhouseid"] == 2
    freezer.tick(UPDATE_SOON)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 6
    assert activity._pending_updates["myhouseid"] == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 7
    assert activity._pending_updates["myhouseid"] == 0
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 7
    assert activity._pending_updates["myhouseid"] == 0
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 7
    assert activity._pending_updates["myhouseid"] == 0
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 7
    assert activity._pending_updates["myhouseid"] == 0


@pytest.mark.asyncio
async def test_activity_stream_debounce_during_init(
    freezer: FrozenDateTimeFactory,
) -> None:
    """Make sure requests during the initial sync get deferred."""

    api = MagicMock(auto_spec=ApiAsync)
    async_get_house_activities = AsyncMock()
    api.async_get_house_activities = async_get_house_activities
    august_gateway = MagicMock(auto_spec=Gateway)
    august_gateway.async_refresh_access_token_if_needed = AsyncMock()
    august_gateway.async_get_access_token = AsyncMock()
    push = MagicMock(connected=False)
    august_gateway.push = push

    activity = ActivityStream(api, august_gateway, {"myhouseid"}, push)
    await activity.async_setup()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)

    assert async_get_house_activities.call_count == 1

    activity.async_schedule_house_id_refresh("myhouseid")
    assert activity._pending_updates["myhouseid"] == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 1

    activity.async_schedule_house_id_refresh("myhouseid")
    assert activity._pending_updates["myhouseid"] == 1
    freezer.tick(ACTIVITY_DEBOUNCE_COOLDOWN + 1)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 1

    freezer.tick(INITIAL_LOCK_RESYNC_TIME)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    assert "myhouseid" not in activity._schedule_updates

    freezer.tick(INITIAL_LOCK_RESYNC_TIME)
    fire_time_changed()
    await asyncio.sleep(0)
    assert async_get_house_activities.call_count == 2
    assert "myhouseid" not in activity._schedule_updates



"""Constants."""

from __future__ import annotations

from datetime import timedelta
from typing import Final

DEFAULT_TIMEOUT = 25

CONF_USERNAME: Final = "username"
CONF_PASSWORD: Final = "password"
CONF_TIMEOUT: Final = "timeout"
CONF_ACCESS_TOKEN_CACHE_FILE: Final = "access_token_cache_file"
CONF_BRAND: Final = "brand"
CONF_LOGIN_METHOD: Final = "login_method"
CONF_INSTALL_ID: Final = "install_id"
VERIFICATION_CODE_KEY: Final = "verification_code"

DEFAULT_AUGUST_CONFIG_FILE: Final = ".august.conf"

# Activity needs to be checked more frequently as the
# doorbell motion and rings are included here
ACTIVITY_UPDATE_INTERVAL = timedelta(seconds=10)


# Limit battery, online, and hardware updates to hourly
# in order to reduce the number of api requests and
# avoid hitting rate limits
MIN_TIME_BETWEEN_DETAIL_UPDATES = timedelta(hours=24)



import time

import pytest

from yalexs.exceptions import RateLimited
from yalexs.manager.ratelimit import RATE_LIMIT_WAKEUP_INTERVAL, _RateLimitChecker


@pytest.mark.asyncio
async def test_init_rate_limit():
    _RateLimitChecker._client_wakeups.clear()

    await _RateLimitChecker.check_rate_limit("test")
    await _RateLimitChecker.register_wakeup("test")
    with pytest.raises(RateLimited) as exc:
        await _RateLimitChecker.check_rate_limit("test")

    assert exc.value.next_allowed == pytest.approx(
        time.monotonic() + RATE_LIMIT_WAKEUP_INTERVAL
    )



"""Handle connection setup and authentication."""

from __future__ import annotations

import asyncio
import logging
import os
from http import HTTPStatus
from pathlib import Path
from typing import TypedDict

from aiohttp import ClientError, ClientResponseError, ClientSession

from ..api_async import ApiAsync
from ..authenticator_async import AuthenticationState, AuthenticatorAsync
from ..authenticator_common import Authentication
from ..const import DEFAULT_BRAND
from ..exceptions import AugustApiAIOHTTPError, RateLimited
from .const import (
    CONF_ACCESS_TOKEN_CACHE_FILE,
    CONF_BRAND,
    CONF_INSTALL_ID,
    CONF_LOGIN_METHOD,
    CONF_PASSWORD,
    CONF_TIMEOUT,
    CONF_USERNAME,
    DEFAULT_AUGUST_CONFIG_FILE,
    DEFAULT_TIMEOUT,
    VERIFICATION_CODE_KEY,
)
from .exceptions import CannotConnect, InvalidAuth, RequireValidation
from .ratelimit import _RateLimitChecker

_LOGGER = logging.getLogger(__name__)


class Config(TypedDict):
    """Config for the gateway."""

    username: str
    password: str
    login_method: str
    access_token_cache_file: str
    install_id: str
    brand: str
    timeout: int


class Gateway:
    """Handle the connection to yale."""

    api: ApiAsync
    authenticator: AuthenticatorAsync
    authentication: Authentication
    _access_token_cache_file: str

    def __init__(self, config_path: Path, aiohttp_session: ClientSession) -> None:
        """Init the connection."""
        self._aiohttp_session = aiohttp_session
        self._token_refresh_lock = asyncio.Lock()
        self._config_path = config_path
        self._config: Config | None = None
        self._loop = asyncio.get_running_loop()

    async def async_get_access_token(self) -> str:
        """Get the access token."""
        return self.authentication.access_token

    def async_configure_access_token_cache_file(
        self, username: str, access_token_cache_file: str | None
    ) -> str:
        """Configure the access token cache file."""
        file = access_token_cache_file or f".{username}{DEFAULT_AUGUST_CONFIG_FILE}"
        self._access_token_cache_file = file
        return self._config_path.joinpath(file)

    async def async_setup(
        self, conf: Config, authenticator_class: type[AuthenticatorAsync] | None = None
    ) -> None:
        """Create the api and authenticator objects."""
        if conf.get(VERIFICATION_CODE_KEY):
            return

        self._config = conf
        self.api = ApiAsync(
            self._aiohttp_session,
            timeout=self._config.get(CONF_TIMEOUT, DEFAULT_TIMEOUT),
            brand=self._config.get(CONF_BRAND, DEFAULT_BRAND),
        )
        klass = authenticator_class or AuthenticatorAsync
        username = conf.get(CONF_USERNAME)
        access_token_cache_file_path: str | None = None
        if username:
            access_token_cache_file_path = self.async_configure_access_token_cache_file(
                conf[CONF_USERNAME], conf.get(CONF_ACCESS_TOKEN_CACHE_FILE)
            )

        self.authenticator = klass(
            self.api,
            self._config.get(CONF_LOGIN_METHOD),
            username,
            self._config.get(CONF_PASSWORD, ""),
            install_id=self._config.get(CONF_INSTALL_ID),
            access_token_cache_file=access_token_cache_file_path,
        )

        await self.authenticator.async_setup_authentication()

    async def async_authenticate(self) -> Authentication:  # noqa: C901
        """Authenticate with the details provided to setup."""
        try:
            self.authentication = await self.authenticator.async_authenticate()
            token = await self.async_get_access_token()
            await _RateLimitChecker.check_rate_limit(token)
            auth_state = self.authentication.state
            if auth_state is AuthenticationState.AUTHENTICATED:
                # Call the locks api to verify we are actually
                # authenticated because we can be authenticated
                # by have no access
                await self.api.async_get_operable_locks(
                    await self.async_get_access_token()
                )
        except RateLimited:
            raise
        except AugustApiAIOHTTPError as ex:
            if ex.auth_failed:
                raise InvalidAuth(ex.args[0], ex.aiohttp_client_error) from ex
            raise CannotConnect(ex.args[0], ex.aiohttp_client_error) from ex
        except ClientResponseError as ex:
            if ex.status == HTTPStatus.UNAUTHORIZED:
                raise InvalidAuth(ex.args[0], ex) from ex

            raise CannotConnect(ex.args[0], ex) from ex
        except ClientError as ex:
            _LOGGER.error("Unable to connect to August service: %s", str(ex))
            raise CannotConnect(ex.args[0], ex) from ex

        if auth_state is AuthenticationState.BAD_PASSWORD:
            raise InvalidAuth

        if auth_state is AuthenticationState.REQUIRES_VALIDATION:
            raise RequireValidation

        if auth_state is not AuthenticationState.AUTHENTICATED:
            _LOGGER.error("Unknown authentication state: %s", auth_state)
            raise InvalidAuth

        return self.authentication

    async def async_reset_authentication(self) -> None:
        """Remove the cache file."""
        await self._loop.run_in_executor(None, self._reset_authentication)

    def _reset_authentication(self) -> None:
        """Remove the cache file."""
        path = self._config_path.joinpath(self._access_token_cache_file)
        if os.path.exists(path):
            os.unlink(path)

    async def async_refresh_access_token_if_needed(self) -> None:
        """Refresh the august access token if needed."""
        if not self.authenticator.should_refresh():
            return
        async with self._token_refresh_lock:
            refreshed_authentication = (
                await self.authenticator.async_refresh_access_token(force=False)
            )
            _LOGGER.info(
                (
                    "Refreshed august access token. The old token expired at %s, and"
                    " the new token expires at %s"
                ),
                self.authentication.access_token_expires,
                refreshed_authentication.access_token_expires,
            )
            self.authentication = refreshed_authentication



"""Device capabilities type definitions."""

from typing import Any, TypedDict


class LockCapabilities(TypedDict, total=False):
    """TypedDict for lock capabilities."""

    concurrentBLE: int
    batteryType: str
    doorSense: bool
    hasMagnetometer: bool
    hasIntegratedWiFi: bool
    scheduledSmartAlerts: bool
    standalone: bool
    bluetooth: bool
    slotRange: Any  # Can be None or other values
    integratedKeypad: bool
    entryCodeSlots: bool
    pinSlotMax: int
    pinSlotMin: int
    supportsRFID: bool
    supportsRFIDLegacy: bool
    supportsRFIDCredential: bool
    supportsRFIDOnlyAccess: bool
    supportsRFIDWithCode: bool
    supportsSecureMode: bool
    supportsSecureModeCodeDisable: bool
    supportsSecureModeMobileControl: bool
    supportsFingerprintCredential: bool
    supportsDeliveryMode: bool
    supportsSchedulePerUser: bool
    supportsFingerprintOnlyAccess: bool
    batteryLifeMS: int
    supportedPartners: list[str]
    unlatch: bool


class CapabilitiesResponse(TypedDict, total=False):
    """TypedDict for the full capabilities API response."""

    lock: LockCapabilities



"""Api functions common between sync and async."""

from __future__ import annotations

import datetime
import logging
from functools import cache
from typing import Any

from ._compat import cached_property
from .activity import ACTION_TO_CLASS, SOURCE_LOCK_OPERATE, SOURCE_LOG, ActivityTypes
from .alarm import Alarm, AlarmDevice, ArmState
from .const import BASE_URLS, BRAND_CONFIG, BRANDING, DEFAULT_BRAND, Brand, BrandConfig
from .doorbell import Doorbell
from .lock import Lock, LockDoorStatus, determine_door_state, door_state_to_string
from .time import parse_datetime

API_EXCEPTION_RETRY_TIME = 0.1
API_RETRY_TIME = 2.5
API_RETRY_ATTEMPTS = 10

HEADER_ACCEPT_VERSION = "Accept-Version"
HEADER_AUGUST_COUNTRY = "x-august-country"
HEADER_CONTENT_TYPE = "Content-Type"
HEADER_USER_AGENT = "User-Agent"


HEADER_VALUE_CONTENT_TYPE = "application/json; charset=UTF-8"
HEADER_VALUE_USER_AGENT = "August/Luna-22.17.0 (Android; SDK 31; gphone64_arm64)"
HEADER_VALUE_ACCEPT_VERSION = "0.0.1"
HEADER_VALUE_AUGUST_BRANDING = "august"
HEADER_VALUE_AUGUST_COUNTRY = "US"


API_GET_SESSION_URL = "/session"
API_SEND_VERIFICATION_CODE_URLS = {
    "phone": "/validation/phone",
    "email": "/validation/email",
}
API_VALIDATE_VERIFICATION_CODE_URLS = {
    "phone": "/validate/phone",
    "email": "/validate/email",
}
API_GET_HOUSE_ACTIVITIES_URL = "/houses/{house_id}/activities"
API_GET_DOORBELLS_URL = "/users/doorbells/mine"
API_GET_DOORBELL_URL = "/doorbells/{doorbell_id}"
API_WAKEUP_DOORBELL_URL = "/doorbells/{doorbell_id}/wakeup"
API_GET_HOUSES_URL = "/users/houses/mine"
API_GET_HOUSE_URL = "/houses/{house_id}"
API_GET_LOCKS_URL = "/users/locks/mine"
API_GET_LOCK_URL = "/locks/{lock_id}"
API_GET_LOCK_STATUS_URL = "/locks/{lock_id}/status"
API_GET_PINS_URL = "/locks/{lock_id}/pins"
API_LOCK_URL = "/remoteoperate/{lock_id}/lock"
API_UNLOCK_URL = "/remoteoperate/{lock_id}/unlock"
API_UNLATCH_URL = "/remoteoperate/{lock_id}/unlatch"
API_LOCK_ASYNC_URL = "/remoteoperate/{lock_id}/lock?v=2.3.1&type=async"
API_UNLOCK_ASYNC_URL = "/remoteoperate/{lock_id}/unlock?v=2.3.1&type=async"
API_UNLATCH_ASYNC_URL = "/remoteoperate/{lock_id}/unlatch?v=2.3.1&type=async"
API_STATUS_ASYNC_URL = (
    "/remoteoperate/{lock_id}/status?v=2.3.1&type=async&intent=wakeup"
)
HYPER_BRIDGE_PARAM = "&connection=persistent"
API_GET_USER_URL = "/users/me"
API_WEBSOCKET_SUBSCRIBERS = "/websocket/subscribers"
API_WEBSOCKET_SUBSCRIBERS_WITH_SUBSCRIBER_ID = "/websocket/subscribers/{subscriber_id}"
API_GET_ALARMS_URL = "/users/alarms/mine"
API_GET_ALARM_DEVICES_URL = "/alarms/{alarm_id}/devices"
API_PUT_ALARM_URL = "/alarms/{alarm_id}/state/{arm_state}"
API_GET_CAPABILITIES_URL = "/devices/capabilities"


_LOGGER = logging.getLogger(__name__)


@cache
def _get_brand_config(brand: Brand) -> BrandConfig:
    return BRAND_CONFIG.get(brand, BRAND_CONFIG[DEFAULT_BRAND])


def api_auth_headers(
    access_token: str | None = None, brand: Brand | None = None
) -> dict[str, str]:
    brand_config = _get_brand_config(brand)
    base_headers = {
        brand_config.api_key_header: brand_config.api_key,
        brand_config.branding_header: BRANDING.get(brand, HEADER_VALUE_AUGUST_BRANDING),
    }
    if access_token:
        base_headers[brand_config.access_token_header] = access_token
    return base_headers


def _api_headers(
    access_token: str | None = None, brand: Brand | None = None
) -> dict[str, str]:
    headers = api_auth_headers(access_token, brand)
    headers.update(
        {
            HEADER_ACCEPT_VERSION: HEADER_VALUE_ACCEPT_VERSION,
            HEADER_CONTENT_TYPE: HEADER_VALUE_CONTENT_TYPE,
            HEADER_AUGUST_COUNTRY: HEADER_VALUE_AUGUST_COUNTRY,
        }
    )
    return headers


def _convert_lock_result_to_activities(
    lock_json_dict: dict[str, Any],
) -> list[ActivityTypes]:
    activities = []
    lock_info_json_dict = lock_json_dict.get("info", {})
    lock_id = lock_info_json_dict.get("lockID")
    lock_action_text = lock_info_json_dict.get("action")
    activity_epoch = _datetime_string_to_epoch(lock_info_json_dict.get("startTime"))
    activity_lock_dict = _map_lock_result_to_activity(
        lock_id, activity_epoch, lock_action_text
    )
    activities.append(activity_lock_dict)

    door_state = determine_door_state(lock_json_dict.get("doorState"))
    if door_state not in (LockDoorStatus.UNKNOWN, LockDoorStatus.DISABLED):
        activity_door_dict = _map_lock_result_to_activity(
            lock_id, activity_epoch, door_state_to_string(door_state)
        )
        activities.append(activity_door_dict)

    return activities


def _activity_from_dict(
    source: str, activity_dict: dict[str, Any], debug: bool = False
) -> ActivityTypes | None:
    """Convert an activity dict to and Activity object."""
    if debug:
        _LOGGER.debug("Processing activity: %s", activity_dict)
    if (action := activity_dict.get("action")) and (
        klass := ACTION_TO_CLASS.get(action)
    ):
        return klass(source, activity_dict)
    if debug:
        _LOGGER.debug("Unknown activity: %s", activity_dict)
    return None


def _map_lock_result_to_activity(
    lock_id: str, activity_epoch: float, action_text: str
) -> ActivityTypes | None:
    """Create an yale access activity from a lock result."""
    mapped_dict = {
        "dateTime": activity_epoch,
        "deviceID": lock_id,
        "deviceType": "lock",
        "action": action_text,
    }
    return _activity_from_dict(
        SOURCE_LOCK_OPERATE, mapped_dict, _LOGGER.isEnabledFor(logging.DEBUG)
    )


def _datetime_string_to_epoch(datetime_string: str) -> datetime.datetime:
    return parse_datetime(datetime_string).timestamp() * 1000


def _process_activity_json(json_dict: dict[str, Any]) -> list[ActivityTypes]:
    if "events" in json_dict:
        json_dict = json_dict["events"]
    debug = _LOGGER.isEnabledFor(logging.DEBUG)
    return [
        activity
        for activity_json in json_dict
        if (activity := _activity_from_dict(SOURCE_LOG, activity_json, debug))
    ]


def _process_doorbells_json(json_dict: dict[str, Any]) -> list[Doorbell]:
    return [Doorbell(device_id, data) for device_id, data in json_dict.items()]


def _process_locks_json(json_dict: dict[str, Any]) -> list[Lock]:
    return [Lock(device_id, data) for device_id, data in json_dict.items()]


def _process_alarms_json(json_dict: list[dict[str, Any]]) -> list[Alarm]:
    return [Alarm(data.get("alarmID"), data) for data in json_dict]


def _process_alarm_devices_json(json_dict: list[dict[str, Any]]) -> list[AlarmDevice]:
    return [AlarmDevice(data) for data in json_dict]


class ApiCommon:
    """Api dict shared between async and sync."""

    def __init__(self, brand: Brand) -> None:
        """Init."""
        self._base_url = BASE_URLS[brand]
        self.brand = brand
        self.brand_config = _get_brand_config(brand)

    @cached_property
    def brand_supports_doorbells(self) -> bool:
        """Return if the brand supports doorbells."""
        return self.brand_config.supports_doorbells

    @cached_property
    def brand_supports_alarms(self) -> bool:
        """Return if the brand supports alarms."""
        return self.brand_config.supports_alarms

    def get_brand_url(self, url_str: str) -> str:
        """Get url."""
        return f"{self._base_url}{url_str}"

    def _build_get_session_request(self, install_id, identifier, password):
        return {
            "method": "post",
            "url": self.get_brand_url(API_GET_SESSION_URL),
            "json": {
                "installId": install_id,
                "identifier": identifier,
                "password": password,
            },
        }

    def _build_send_verification_code_request(
        self, access_token, login_method, username
    ):
        if login_method == "phone":
            json = {"smsHashString": "anY0ZsRmXw+", "value": username}
        else:
            json = {"value": username}

        return {
            **self._build_base_request(access_token, "post"),
            "url": self.get_brand_url(API_SEND_VERIFICATION_CODE_URLS[login_method]),
            "json": json,
        }

    def _build_base_request(
        self, access_token: str, method: str = "get"
    ) -> dict[str, Any]:
        """Build a base request."""
        return {"method": method, "access_token": access_token}

    def _build_validate_verification_code_request(
        self, access_token, login_method, username, verification_code
    ):
        return {
            **self._build_base_request(access_token, "post"),
            "url": self.get_brand_url(
                API_VALIDATE_VERIFICATION_CODE_URLS[login_method]
            ),
            "json": {login_method: username, "code": str(verification_code)},
        }

    def _build_get_doorbells_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_DOORBELLS_URL),
        }

    def _build_get_doorbell_detail_request(
        self, access_token: str, doorbell_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(
                API_GET_DOORBELL_URL.format(doorbell_id=doorbell_id)
            ),
        }

    def _build_wakeup_doorbell_request(
        self, access_token: str, doorbell_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(
                API_WAKEUP_DOORBELL_URL.format(doorbell_id=doorbell_id)
            ),
        }

    def _build_get_houses_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_HOUSES_URL),
        }

    def _build_get_house_request(self, access_token, house_id):
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_HOUSE_URL.format(house_id=house_id)),
        }

    def _build_get_house_activities_request(self, access_token, house_id, limit=8):
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(
                API_GET_HOUSE_ACTIVITIES_URL.format(house_id=house_id)
            ),
            "version": "4.0.0",
            "params": {"limit": limit},
        }

    def _build_get_locks_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_LOCKS_URL),
        }

    def _build_get_user_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_USER_URL),
        }

    def _build_get_lock_detail_request(
        self, access_token: str, lock_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_LOCK_URL.format(lock_id=lock_id)),
        }

    def _build_get_lock_status_request(
        self, access_token: str, lock_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_LOCK_STATUS_URL.format(lock_id=lock_id)),
        }

    def _build_get_pins_request(
        self, access_token: str, lock_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_PINS_URL.format(lock_id=lock_id)),
        }

    def _build_get_capabilities_request(
        self, access_token: str, serial_number: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_CAPABILITIES_URL),
            "params": {"serialNumber": serial_number, "topLevelHost": "true"},
        }

    def _build_refresh_access_token_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_HOUSES_URL),
        }

    def _build_websocket_subscribe_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token, "post"),
            "url": self.get_brand_url(API_WEBSOCKET_SUBSCRIBERS),
            "json": {
                "scopes": ["lock"],
            },
        }

    def _build_websocket_get_request(
        self, access_token: str, subscriber_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token, "get"),
            "url": self.get_brand_url(
                API_WEBSOCKET_SUBSCRIBERS_WITH_SUBSCRIBER_ID.format(
                    subscriber_id=subscriber_id
                )
            ),
        }

    def _build_websocket_delete_request(
        self, access_token: str, subscriber_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token, "delete"),
            "url": self.get_brand_url(
                API_WEBSOCKET_SUBSCRIBERS_WITH_SUBSCRIBER_ID.format(
                    subscriber_id=subscriber_id
                )
            ),
        }

    def _build_call_lock_operation_request(
        self, url_str: str, access_token: str, lock_id: str, timeout
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token, "put"),
            "url": self.get_brand_url(url_str.format(lock_id=lock_id)),
            "timeout": timeout,
        }

    def _build_get_alarms_request(self, access_token: str) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(API_GET_ALARMS_URL),
        }

    def _build_get_alarm_devices_request(
        self, access_token: str, alarm_id: str
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token),
            "url": self.get_brand_url(
                API_GET_ALARM_DEVICES_URL.format(alarm_id=alarm_id)
            ),
        }

    def _build_call_alarm_state_request(
        self, access_token: str, alarm: Alarm, arm_state: ArmState
    ) -> dict[str, Any]:
        return {
            **self._build_base_request(access_token=access_token, method="PUT"),
            "url": self.get_brand_url(
                API_PUT_ALARM_URL.format(alarm_id=alarm.device_id, arm_state=arm_state)
            ),
            "json": {"areaIDs": alarm.areaIDs},
        }



from __future__ import annotations

from typing import Any

from ._compat import cached_property

USER_CACHE = {}


class YaleUser:
    """Represent a yale access user."""

    def __init__(self, uuid: str, data: dict[str, Any]) -> None:
        """Initialize the YaleUser."""
        self._uuid = uuid
        self._data = data

    @cached_property
    def thumbnail_url(self) -> str | None:
        return self._data.get("imageInfo", {}).get("thumbnail", {}).get("secure_url")

    @cached_property
    def image_url(self) -> str | None:
        return self._data.get("imageInfo", {}).get("original", {}).get("secure_url")

    @cached_property
    def first_name(self) -> str | None:
        return self._data.get("FirstName")

    @cached_property
    def last_name(self) -> str | None:
        return self._data.get("LastName")

    @cached_property
    def user_type(self) -> str | None:
        return self._data.get("UserType")


def get_user_info(uuid: str) -> YaleUser | None:
    return USER_CACHE.get(uuid)


def cache_user_info(uuid: str, data: dict[str, Any]) -> None:
    if uuid not in USER_CACHE:
        USER_CACHE[uuid] = YaleUser(uuid, data)



"""yalexs subscribers."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable
from datetime import timedelta
from functools import partial
from typing import Any

from ..backports.tasks import create_eager_task


class SubscriberMixin(ABC):
    """Base implementation for a subscriber."""

    def __init__(self, update_interval: timedelta) -> None:
        """Initialize an subscriber."""
        super().__init__()
        self._update_interval_seconds = update_interval.total_seconds()
        self._subscriptions: defaultdict[str, set[Callable[[], None]]] = defaultdict(
            set
        )
        self._unsub_interval: asyncio.TimerHandle | None = None
        self._loop = asyncio.get_running_loop()
        self._refresh_task: asyncio.Task | None = None

    def async_subscribe_device_id(
        self, device_id: str, update_callback: Callable[[], None]
    ) -> Callable[[], None]:
        """Add an callback subscriber.

        Returns a callable that can be used to unsubscribe.
        """
        if not self._subscriptions:
            self._async_setup_listeners()
        self._subscriptions[device_id].add(update_callback)
        return partial(self.async_unsubscribe_device_id, device_id, update_callback)

    @abstractmethod
    async def _async_refresh(self) -> None:
        """Refresh data."""

    def _async_scheduled_refresh(self) -> None:
        """Call the refresh method."""
        self._unsub_interval = self._loop.call_later(
            self._update_interval_seconds,
            self._async_scheduled_refresh,
        )
        self._refresh_task = create_eager_task(
            self._async_refresh(), loop=self._loop, name=f"{self} schedule refresh"
        )

    def _async_cancel_update_interval(self) -> None:
        """Cancel the scheduled update."""
        if self._unsub_interval:
            self._unsub_interval.cancel()
            self._unsub_interval = None

    def _async_setup_listeners(self) -> None:
        """Create interval and stop listeners."""
        self._async_cancel_update_interval()
        self._unsub_interval = self._loop.call_later(
            self._update_interval_seconds,
            self._async_scheduled_refresh,
        )

    def async_stop(self, *args: Any) -> None:
        """Cleanup on shutdown."""
        self._refresh_task.cancel()
        self._async_cancel_update_interval()

    def async_unsubscribe_device_id(
        self, device_id: str, update_callback: Callable[[], None]
    ) -> None:
        """Remove a callback subscriber."""
        self._subscriptions[device_id].remove(update_callback)
        if not self._subscriptions[device_id]:
            del self._subscriptions[device_id]
        if self._subscriptions:
            return
        self._async_cancel_update_interval()

    def async_signal_device_id_update(self, device_id: str) -> None:
        """Call the callbacks for a device_id."""
        for update_callback in self._subscriptions.get(device_id, ()):
            update_callback()



import datetime
import json
import os
import unittest

import dateutil.parser
from dateutil.tz import tzlocal

from yalexs.activity import (
    SOURCE_PUBNUB,
    SOURCE_WEBSOCKET,
    ActivityType,
    BridgeOperationActivity,
    DoorbellDingActivity,
    DoorbellImageCaptureActivity,
    DoorbellMotionActivity,
    DoorOperationActivity,
    LockOperationActivity,
)
from yalexs.doorbell import DoorbellDetail
from yalexs.lock import (
    DOOR_STATE_KEY,
    LOCK_STATUS_KEY,
    LockDetail,
    LockDoorStatus,
    LockStatus,
)
from yalexs.pubnub_activity import activities_from_pubnub_message
from yalexs.users import cache_user_info, get_user_info


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path) as fptr:
        return fptr.read()


class TestLockDetail(unittest.TestCase):
    def test_update_lock_details_from_pubnub_message(self):
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
        self.assertEqual("A6697750D607098BAE8D6BAA11EF8063", lock.device_id)
        self.assertEqual(LockStatus.LOCKED, lock.lock_status)
        self.assertEqual(LockDoorStatus.DISABLED, lock.door_state)

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "remoteEvent": 1,
                "status": "kAugLockState_Unlatching",
                "info": {
                    "action": "unlock",
                    "startTime": "2021-03-20T18:19:05.373Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2021-03-20T18:19:05.371Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "unlatching"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "remoteEvent": 1,
                "status": "kAugLockState_Unlocking",
                "info": {
                    "action": "unlock",
                    "startTime": "2021-03-20T18:19:05.373Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2021-03-20T18:19:05.371Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "unlocking"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:31.273Z"),
            {
                "remoteEvent": 1,
                "status": "kAugLockState_Locking",
                "info": {
                    "action": "unlock",
                    "startTime": "2021-03-20T18:19:06.374Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2021-03-20T18:19:06.372Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "locking"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:31.273Z"),
            {
                "remoteEvent": 1,
                "status": "FAILED_BRIDGE_ERROR_LOCK_JAMMED",
                "info": {
                    "action": "unlock",
                    "startTime": "2021-03-20T18:19:06.374Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2021-03-20T18:19:06.372Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert activities[0].activity_start_time == dateutil.parser.parse(
            "2021-03-20T18:19:06.374Z"
        ).astimezone(tz=tzlocal()).replace(tzinfo=None)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "jammed"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:31.273Z"),
            {
                "remoteEvent": 1,
                "status": "kAugLockState_Unlocked",
                "info": {
                    "action": "unlock",
                    "startTime": "2021-03-20T18:19:06.374Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2021-03-20T18:19:06.372Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].activity_start_time == dateutil.parser.parse(
            "2021-03-20T18:19:06.374Z"
        ).astimezone(tz=tzlocal()).replace(tzinfo=None)
        assert activities[0].action == "unlock"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "locked",
                "callingUserID": "8918341e-7e68-4079-ad0a-1fa8a45d855b",
                "doorState": "closed",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "lock"
        assert activities[0].operated_by is None
        assert isinstance(activities[1], DoorOperationActivity)
        assert "DoorOperationActivity" in str(activities[1])
        assert activities[1].action == "doorclosed"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "locked",
                "callingUserID": "8918341e-7e68-4079-ad0a-1fa8a45d855b",
                "doorState": "open",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert activities[0].action == "lock"
        assert activities[0].operated_by is None
        assert (
            activities[0].activity_type == ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR
        )
        assert isinstance(activities[1], DoorOperationActivity)
        assert activities[1].action == "dooropen"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "locked",
                "callingUserID": "cccca94e-373e-aaaa-bbbb-333396827777",
                "doorState": "closed",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert activities[0].action == "lock"
        assert activities[0].operated_by is None
        assert (
            activities[0].activity_type == ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR
        )
        assert isinstance(activities[1], DoorOperationActivity)
        assert activities[1].action == "doorclosed"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T11:48:30.272Z"),
            {
                LOCK_STATUS_KEY: "DoorStateChanged",
                "lockID": "xxx",
                "timeStamp": 1615087688187,
            },
        )
        assert not activities
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T12:48:30.272Z"),
            {
                DOOR_STATE_KEY: "init",
                "lockID": "xxx",
                "timeStamp": 1615087688187,
            },
        )
        assert not activities

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T11:48:30.272Z"),
            {
                LOCK_STATUS_KEY: "associated_bridge_offline",
                "lockID": "xxx",
                "timeStamp": 1615087688187,
            },
        )
        assert isinstance(activities[0], BridgeOperationActivity)
        assert activities[0].action == "associated_bridge_offline"

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T11:48:30.272Z"),
            {
                LOCK_STATUS_KEY: "associated_bridge_online",
                "lockID": "xxx",
                "timeStamp": 1615087688187,
            },
        )
        assert isinstance(activities[0], BridgeOperationActivity)
        assert activities[0].action == "associated_bridge_online"

        cache_user_info(
            "5309b78d-de0c-4ec5-b878-02784c3b598a",
            {"FirstName": "bob", "LastName": "smith"},
        )
        assert get_user_info("5309b78d-de0c-4ec5-b878-02784c3b598a") is not None

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "unlocked",
                "callingUserID": "5309b78d-de0c-4ec5-b878-02784c3b598a",
                "doorState": "closed",
                "info": {
                    "action": "unlock",
                    "startTime": "2017-12-10T05:48:30.272Z",
                    "context": {
                        "transactionID": "_oJRZKJsx",
                        "startDate": "2017-12-10T05:48:30.272Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_1001",
                    "serialNumber": "M1FBA029QJ",
                    "rssi": -53,
                    "wlanRSSI": -55,
                    "wlanSNR": 44,
                    "duration": 2534,
                },
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "unlock"
        assert activities[0].operated_by == "bob smith"

        activities = activities_from_pubnub_message(
            lock,
            datetime.datetime.fromtimestamp(16844292526891571 / 10_000_000),
            {
                "status": "unlatched",
                "callingUserID": "manualunlatch",
                "doorState": "open",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "unlatch"
        assert (
            activities[0].activity_type is ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR
        )
        assert activities[0].operated_by is None

        activities = activities_from_pubnub_message(
            lock,
            datetime.datetime.fromtimestamp(16844292526891571 / 10_000_000),
            {
                "status": "unlocked",
                "callingUserID": "manualunlock",
                "doorState": "open",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "unlock"
        assert (
            activities[0].activity_type is ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR
        )
        assert activities[0].operated_by is None

        activities = activities_from_pubnub_message(
            lock,
            datetime.datetime.fromtimestamp(16844299539729015 / 10_000_000),
            {
                "status": "locked",
                "callingUserID": "manuallock",
                "doorState": "open",
            },
        )
        assert isinstance(activities[0], LockOperationActivity)
        assert "LockOperationActivity" in str(activities[0])
        assert activities[0].action == "lock"
        assert (
            activities[0].activity_type is ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR
        )
        assert activities[0].operated_by is None

        # status polls should not create activities
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "remoteEvent": 1,
                "status": "kAugLockState_Locked",
                "info": {
                    "action": "status",
                    "startTime": "2024-02-15T07:33:50.804Z",
                    "context": {
                        "transactionID": "RP99lHGUIx",
                        "startDate": "2024-02-15T07:33:50.793Z",
                        "retryCount": 1,
                    },
                    "lockType": "lock_version_17",
                    "serialNumber": "L.....",
                    "rssi": 0,
                    "wlanRSSI": -35,
                    "wlanSNR": -1,
                    "duration": 991,
                    "lockID": "AF5EFD84.....",
                    "bridgeID": "652e35ba7e.....",
                    "serial": "L.....",
                },
                "doorState": "kAugDoorState_Closed",
                "retryCount": 1,
                "totalTime": 1028,
                "resultsFromOperationCache": False,
            },
        )

        assert len(activities) == 0


class TestDetail(unittest.TestCase):
    def test_update_doorbell_details_from_pubnub_message(self):
        doorbell = DoorbellDetail(json.loads(load_fixture("get_doorbell.json")))
        self.assertEqual("K98GiDT45GUL", doorbell.device_id)
        self.assertEqual(
            dateutil.parser.parse("2017-12-10T08:01:35Z"),
            doorbell.image_created_at_datetime,
        )
        self.assertEqual(
            "https://image.com/vmk16naaaa7ibuey7sar.jpg", doorbell.image_url
        )
        activities = activities_from_pubnub_message(
            doorbell,
            dateutil.parser.parse("2021-03-16T01:07:08.817Z"),
            {
                "status": "imagecapture",
                "data": {
                    "event": "imagecapture",
                    "result": {
                        "created_at": "2021-03-16T01:07:08.817Z",
                        "secure_url": "https://dyu7azbnaoi74.cloudfront.net/zip/images/zip.jpeg",
                    },
                },
            },
        )
        assert isinstance(activities[0], DoorbellImageCaptureActivity)
        assert "DoorbellImageCaptureActivity" in str(activities[0])

        assert (
            activities[0].image_url
            == "https://dyu7azbnaoi74.cloudfront.net/zip/images/zip.jpeg"
        )
        assert activities[0].image_created_at_datetime == dateutil.parser.parse(
            "2021-03-16T01:07:08.817Z"
        )

        activities = activities_from_pubnub_message(
            doorbell,
            dateutil.parser.parse("2021-03-16T01:07:08.817Z"),
            {
                "status": "imagecapture",
                "data": {
                    "event": "imagecapture",
                    "result": {
                        "created_at": "2021-03-16T01:07:08.817Z",
                        "secure_url": "https://dyu7azbnaoi74.cloudfront.net/zip/images/zip.jpeg",
                    },
                },
            },
        )
        assert isinstance(activities[0], DoorbellImageCaptureActivity)
        assert (
            activities[0].image_url
            == "https://dyu7azbnaoi74.cloudfront.net/zip/images/zip.jpeg"
        )
        assert activities[0].image_created_at_datetime == dateutil.parser.parse(
            "2021-03-16T01:07:08.817Z"
        )

        activities = activities_from_pubnub_message(
            doorbell,
            dateutil.parser.parse("2021-03-16T01:07:08.817Z"),
            {
                "status": "doorbell_motion_detected",
                "callID": None,
                "origin": "mars-api",
                "data": {
                    "event": "doorbell_motion_detected",
                    "image": {
                        "height": 640,
                        "width": 480,
                        "format": "jpg",
                        "created_at": "2021-03-16T02:36:26.886Z",
                        "bytes": 14061,
                        "secure_url": "https://dyu7azbnaoi74.cloudfront.net/images/1f8.jpeg",
                        "url": "https://dyu7azbnaoi74.cloudfront.net/images/1f8.jpeg",
                        "etag": "09e839331c4ea59eef28081f2caa0e90",
                    },
                    "doorbellName": "Front Door",
                    "callID": None,
                    "origin": "mars-api",
                    "mutableContent": True,
                },
            },
        )
        assert isinstance(activities[0], DoorbellMotionActivity)
        assert (
            activities[0].image_url
            == "https://dyu7azbnaoi74.cloudfront.net/images/1f8.jpeg"
        )
        assert activities[0].image_created_at_datetime == dateutil.parser.parse(
            "2021-03-16T02:36:26.886Z"
        )

        activities = activities_from_pubnub_message(
            doorbell,
            dateutil.parser.parse("2021-03-16T01:07:08.817Z"),
            {
                "status": "buttonpush",
                "origin": "mars-api",
                "data": {
                    "doorbellID": "26593a60f5d6",
                    "event": "buttonpush",
                    "doorbellName": "Front Door",
                    "origin": "mars-api",
                },
            },
        )
        assert isinstance(activities[0], DoorbellDingActivity)
        assert "DoorbellDingActivity" in str(activities[0])


class TestBridge(unittest.TestCase):
    def test_update_bridge_details_from_pubnub_message(self):
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
        self.assertEqual("A6697750D607098BAE8D6BAA11EF8063", lock.device_id)
        self.assertEqual(LockStatus.LOCKED, lock.lock_status)
        self.assertEqual(LockDoorStatus.DISABLED, lock.door_state)

        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "remoteEvent": 1,
                "status": "unknown",
                "result": "failed",
                "error": {
                    "jse_shortmsg": "",
                    "jse_info": {},
                    "message": "Bridge is offline",
                    "statusCode": 422,
                    "body": {"code": 98, "message": "Bridge is offline"},
                    "restCode": 98,
                    "name": "ERRNO_BRIDGE_OFFLINE",
                    "code": "Error",
                },
                "info": {
                    "lockID": "45E3635D35B9471FAF1218885816E90D",
                    "action": "status",
                },
            },
        )
        assert isinstance(activities[0], BridgeOperationActivity)
        assert "BridgeOperationActivity" in str(activities[0])
        assert activities[0].action == "associated_bridge_offline"

    def test_websocket_source_handling(self):
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {"status": "kAugLockState_Locked"},
            source=SOURCE_WEBSOCKET,
        )
        assert len(activities) == 1
        assert activities[0].source == SOURCE_WEBSOCKET

    def test_pubnub_source_default(self):
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {"status": "kAugLockState_Locked"},
        )
        assert len(activities) == 1
        assert activities[0].source == SOURCE_PUBNUB

    def test_is_status_property(self):
        """Test is_status property correctly identifies status updates."""
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))

        # Test normal lock operation is not a status
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "kAugLockState_Locked",
                "info": {"action": "lock", "startTime": "2017-12-10T05:48:30.272Z"},
                "callingUserID": "user123",
            },
        )
        assert len(activities) == 1
        assert activities[0].is_status is False

        # Test status update with no info and no user
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {"status": "kAugLockState_Locked"},
        )
        assert len(activities) == 1
        assert activities[0].is_status is True

        # Test manual operation is not a status
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "kAugLockState_Locked",
                "callingUserID": "manuallock",
            },
        )
        assert len(activities) == 1
        assert activities[0].is_status is False
        # Manual operations without timestamps don't store the user ID but mark as manual
        assert activities[0].operated_manual is True

        # Test WebSocket messages with empty info are not status
        activities = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {"status": "kAugLockState_Locked"},
            source=SOURCE_WEBSOCKET,
        )
        assert len(activities) == 1
        assert activities[0].is_status is False

    def test_pubnub_duplicate_state_detection(self):
        """Test that duplicate PubNub messages with same state are detected."""
        lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))

        # First message - manual lock when already locked
        activities1 = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
            {
                "status": "locked",
                "callingUserID": "manuallock",
                "doorState": "closed",
            },
            source=SOURCE_PUBNUB,
        )
        assert len(activities1) == 2  # lock and door activities
        assert activities1[0].action == "lock"
        assert activities1[1].action == "doorclosed"

        # Second identical message should still create activities
        # (state tracking happens at the manager level, not here)
        activities2 = activities_from_pubnub_message(
            lock,
            dateutil.parser.parse("2017-12-10T05:48:31.272Z"),
            {
                "status": "locked",
                "callingUserID": "manuallock",
                "doorState": "closed",
            },
            source=SOURCE_PUBNUB,
        )
        assert len(activities2) == 2
        assert activities2[0].action == "lock"
        assert activities2[1].action == "doorclosed"
        # Both should NOT be marked as status since they have manual flag
        assert (
            activities2[0].is_status is False
        )  # manual lock operations are not status
        assert (
            activities2[1].is_status is False
        )  # door activity with manual flag is not status



"""Api calls for sync."""

from __future__ import annotations

import asyncio
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .capabilities import CapabilitiesResponse

from aiohttp import (
    ClientConnectionError,
    ClientOSError,
    ClientResponse,
    ClientResponseError,
    ClientSession,
    ClientSSLError,
    ServerDisconnectedError,
)

from .activity import ActivityTypes
from .alarm import Alarm, AlarmDevice, ArmState
from .api_common import (
    API_EXCEPTION_RETRY_TIME,
    API_LOCK_ASYNC_URL,
    API_LOCK_URL,
    API_RETRY_ATTEMPTS,
    API_RETRY_TIME,
    API_STATUS_ASYNC_URL,
    API_UNLATCH_ASYNC_URL,
    API_UNLATCH_URL,
    API_UNLOCK_ASYNC_URL,
    API_UNLOCK_URL,
    HEADER_ACCEPT_VERSION,
    HYPER_BRIDGE_PARAM,
    ApiCommon,
    _api_headers,
    _convert_lock_result_to_activities,
    _process_activity_json,
    _process_alarm_devices_json,
    _process_alarms_json,
    _process_doorbells_json,
    _process_locks_json,
)
from .const import DEFAULT_BRAND, HEADER_ACCESS_TOKEN, HEADER_AUGUST_ACCESS_TOKEN
from .doorbell import Doorbell, DoorbellDetail
from .exceptions import InvalidAuth, YaleApiError
from .lock import (
    Lock,
    LockDetail,
    LockDoorStatus,
    LockStatus,
    determine_door_state,
    determine_lock_status,
)
from .pin import Pin

_LOGGER = logging.getLogger(__name__)


def _obscure_payload(payload: dict[str, Any]) -> dict[str, Any]:
    """Obscure the payload for logging."""
    if payload is None:
        return None
    if "password" in payload:
        payload = payload.copy()
        payload["password"] = "****"  # nosec
    return payload


def _obscure_headers(headers: dict[str, Any]) -> dict[str, Any]:
    """Obscure the headers for logging."""
    if headers is None:
        return None
    for obscure_header in (
        "x-august-access-token",
        "x-access-token",
        "x-august-api-key",
        "x-api-key",
    ):
        if obscure_header in headers:
            headers = headers.copy()
            headers[obscure_header] = "****"
    return headers


class ApiAsync(ApiCommon):
    """Async api."""

    def __init__(
        self,
        aiohttp_session: ClientSession,
        timeout=10,
        command_timeout=60,
        brand=DEFAULT_BRAND,
    ) -> None:
        self._timeout = timeout
        self._command_timeout = command_timeout
        self._aiohttp_session = aiohttp_session
        super().__init__(brand)

    async def async_get_session(
        self, install_id: str, identifier: str, password: str
    ) -> ClientResponse:
        return await self._async_dict_to_api(
            self._build_get_session_request(install_id, identifier, password)
        )

    async def async_send_verification_code(
        self, access_token: str, login_method: str, username: str
    ) -> ClientResponse:
        return await self._async_dict_to_api(
            self._build_send_verification_code_request(
                access_token, login_method, username
            )
        )

    async def async_validate_verification_code(
        self,
        access_token: str,
        login_method: str,
        username: str,
        verification_code: str,
    ) -> ClientResponse:
        return await self._async_dict_to_api(
            self._build_validate_verification_code_request(
                access_token, login_method, username, verification_code
            )
        )

    async def async_get_doorbells(self, access_token: str) -> list[Doorbell]:
        if not self.brand_supports_doorbells:
            return []
        response = await self._async_dict_to_api(
            self._build_get_doorbells_request(access_token)
        )
        return _process_doorbells_json(await response.json())

    async def async_get_doorbell_detail(
        self, access_token: str, doorbell_id: str
    ) -> DoorbellDetail:
        response = await self._async_dict_to_api(
            self._build_get_doorbell_detail_request(access_token, doorbell_id)
        )
        return DoorbellDetail(await response.json())

    async def async_wakeup_doorbell(
        self, access_token: str, doorbell_id: str
    ) -> ClientResponse:
        await self._async_dict_to_api(
            self._build_wakeup_doorbell_request(access_token, doorbell_id)
        )
        return True

    async def async_get_user(self, access_token: str) -> dict[str, Any]:
        response = await self._async_dict_to_api(
            self._build_get_user_request(access_token)
        )
        return await response.json()

    async def async_get_houses(self, access_token: str) -> ClientResponse:
        return await self._async_dict_to_api(
            self._build_get_houses_request(access_token)
        )

    async def async_get_house(self, access_token: str, house_id: str) -> dict[str, Any]:
        response = await self._async_dict_to_api(
            self._build_get_house_request(access_token, house_id)
        )
        return await response.json()

    async def async_get_house_activities(
        self, access_token: str, house_id: str, limit: int = 8
    ) -> list[ActivityTypes]:
        response = await self._async_dict_to_api(
            self._build_get_house_activities_request(
                access_token, house_id, limit=limit
            )
        )
        return _process_activity_json(await response.json())

    async def async_get_locks(self, access_token: str) -> list[Lock]:
        response = await self._async_dict_to_api(
            self._build_get_locks_request(access_token)
        )
        return _process_locks_json(await response.json())

    async def async_get_operable_locks(self, access_token: str) -> list[Lock]:
        locks = await self.async_get_locks(access_token)

        return [lock for lock in locks if lock.is_operable]

    async def async_get_lock_detail(
        self, access_token: str, lock_id: str
    ) -> LockDetail:
        response = await self._async_dict_to_api(
            self._build_get_lock_detail_request(access_token, lock_id)
        )
        return LockDetail(await response.json())

    async def async_get_lock_status(
        self, access_token: str, lock_id: str, door_status=False
    ) -> LockStatus:
        response = await self._async_dict_to_api(
            self._build_get_lock_status_request(access_token, lock_id)
        )
        json_dict = await response.json()

        if door_status:
            return (
                determine_lock_status(json_dict.get("status")),
                determine_door_state(json_dict.get("doorState")),
            )

        return determine_lock_status(json_dict.get("status"))

    async def async_get_lock_door_status(
        self, access_token: str, lock_id: str, lock_status=False
    ) -> LockDoorStatus | tuple[LockDoorStatus, LockStatus]:
        response = await self._async_dict_to_api(
            self._build_get_lock_status_request(access_token, lock_id)
        )
        json_dict = await response.json()

        if lock_status:
            return (
                determine_door_state(json_dict.get("doorState")),
                determine_lock_status(json_dict.get("status")),
            )

        return determine_door_state(json_dict.get("doorState"))

    async def async_get_pins(self, access_token: str, lock_id: str) -> list[Pin]:
        response = await self._async_dict_to_api(
            self._build_get_pins_request(access_token, lock_id)
        )
        json_dict = await response.json()

        return [Pin(pin_json) for pin_json in json_dict.get("loaded", [])]

    async def async_get_lock_capabilities(
        self, access_token: str, serial_number: str
    ) -> CapabilitiesResponse:
        response = await self._async_dict_to_api(
            self._build_get_capabilities_request(access_token, serial_number)
        )
        return await response.json()

    async def _async_call_lock_operation(
        self, url_str: str, access_token: str, lock_id: str
    ) -> dict[str, Any]:
        response = await self._async_dict_to_api(
            self._build_call_lock_operation_request(
                url_str, access_token, lock_id, self._command_timeout
            )
        )
        return await response.json()

    async def _async_call_async_lock_operation(
        self, url_str: str, access_token: str, lock_id: str
    ) -> str:
        """Call an operation that will queue."""
        response = await self._async_dict_to_api(
            self._build_call_lock_operation_request(
                url_str, access_token, lock_id, self._command_timeout
            )
        )
        return await response.text()

    async def _async_lock(self, access_token: str, lock_id: str) -> str:
        return await self._async_call_lock_operation(
            API_LOCK_URL, access_token, lock_id
        )

    async def async_lock(self, access_token: str, lock_id: str) -> str:
        """Execute a remote lock operation.

        Returns a LockStatus state.
        """
        return determine_lock_status(
            (await self._async_lock(access_token, lock_id)).get("status")
        )

    async def async_lock_async(
        self, access_token: str, lock_id: str, hyper_bridge=True
    ) -> str:
        """Queue a remote lock operation and get the response via pubnub."""
        if hyper_bridge:
            return await self._async_call_async_lock_operation(
                f"{API_LOCK_ASYNC_URL}{HYPER_BRIDGE_PARAM}", access_token, lock_id
            )
        return await self._async_call_async_lock_operation(
            API_LOCK_ASYNC_URL, access_token, lock_id
        )

    async def async_lock_return_activities(
        self, access_token: str, lock_id: str
    ) -> list[ActivityTypes]:
        """Execute a remote lock operation.

        Returns an array of one or more yalexs.activity.Activity objects

        If the lock supports door sense one of the activities
        will include the current door state.
        """
        return _convert_lock_result_to_activities(
            await self._async_lock(access_token, lock_id)
        )

    async def _async_unlatch(self, access_token: str, lock_id: str) -> dict[str, Any]:
        return await self._async_call_lock_operation(
            API_UNLATCH_URL, access_token, lock_id
        )

    async def async_unlatch(self, access_token: str, lock_id: str) -> LockStatus:
        """Execute a remote unlatch operation.

        Returns a LockStatus state.
        """
        return determine_lock_status(
            (await self._async_unlatch(access_token, lock_id)).get("status")
        )

    async def async_unlatch_async(
        self, access_token: str, lock_id: str, hyper_bridge=True
    ) -> str:
        """Queue a remote unlatch operation and get the response via pubnub."""
        if hyper_bridge:
            return await self._async_call_async_lock_operation(
                f"{API_UNLATCH_ASYNC_URL}{HYPER_BRIDGE_PARAM}", access_token, lock_id
            )
        return await self._async_call_async_lock_operation(
            API_UNLATCH_ASYNC_URL, access_token, lock_id
        )

    async def async_unlatch_return_activities(
        self, access_token: str, lock_id: str
    ) -> list[ActivityTypes]:
        """Execute a remote lock operation.

        Returns an array of one or more yalexs.activity.Activity objects

        If the lock supports door sense one of the activities
        will include the current door state.
        """
        return _convert_lock_result_to_activities(
            await self._async_unlatch(access_token, lock_id)
        )

    async def _async_unlock(self, access_token: str, lock_id: str) -> dict[str, Any]:
        return await self._async_call_lock_operation(
            API_UNLOCK_URL, access_token, lock_id
        )

    async def async_unlock(self, access_token: str, lock_id: str) -> LockStatus:
        """Execute a remote unlock operation.

        Returns a LockStatus state.
        """
        return determine_lock_status(
            (await self._async_unlock(access_token, lock_id)).get("status")
        )

    async def async_unlock_async(
        self, access_token: str, lock_id: str, hyper_bridge=True
    ) -> str:
        """Queue a remote unlock operation and get the response via pubnub."""
        if hyper_bridge:
            return await self._async_call_async_lock_operation(
                f"{API_UNLOCK_ASYNC_URL}{HYPER_BRIDGE_PARAM}", access_token, lock_id
            )
        return await self._async_call_async_lock_operation(
            API_UNLOCK_ASYNC_URL, access_token, lock_id
        )

    async def async_unlock_return_activities(
        self, access_token: str, lock_id: str
    ) -> list[ActivityTypes]:
        """Execute a remote lock operation.

        Returns an array of one or more yalexs.activity.Activity objects

        If the lock supports door sense one of the activities
        will include the current door state.
        """
        return _convert_lock_result_to_activities(
            await self._async_unlock(access_token, lock_id)
        )

    async def async_status_async(
        self, access_token: str, lock_id: str, hyper_bridge=True
    ) -> str:
        """Queue a remote unlock operation and get the status via pubnub."""
        if hyper_bridge:
            return await self._async_call_async_lock_operation(
                f"{API_STATUS_ASYNC_URL}{HYPER_BRIDGE_PARAM}", access_token, lock_id
            )
        return await self._async_call_async_lock_operation(
            API_STATUS_ASYNC_URL, access_token, lock_id
        )

    async def async_get_alarms(self, access_token: str) -> list[Alarm]:
        if not self.brand_supports_alarms:
            return []
        response = await self._async_dict_to_api(
            self._build_get_alarms_request(access_token)
        )
        return _process_alarms_json(await response.json())

    async def async_get_alarm_devices(
        self, access_token: str, alarm: Alarm
    ) -> list[AlarmDevice]:
        if not self.brand_supports_alarms:
            return []
        response = await self._async_dict_to_api(
            self._build_get_alarm_devices_request(
                access_token, alarm_id=alarm.device_id
            )
        )
        return _process_alarm_devices_json(await response.json())

    async def async_arm_alarm(
        self, access_token: str, alarm: Alarm, arm_state: ArmState
    ):
        if not self.brand_supports_alarms:
            return {}
        response = await self._async_dict_to_api(
            self._build_call_alarm_state_request(access_token, alarm, arm_state)
        )
        return await response.json()

    async def async_refresh_access_token(self, access_token: str) -> str:
        """Obtain a new api token."""
        response = await self._async_dict_to_api(
            self._build_refresh_access_token_request(access_token)
        )
        response_headers = response.headers
        return (
            response_headers.get(HEADER_ACCESS_TOKEN)
            or response_headers[HEADER_AUGUST_ACCESS_TOKEN]
        )

    async def async_add_websocket_subscription(
        self, access_token: str
    ) -> dict[str, Any]:
        """Add a websocket subscription."""
        response = await self._async_dict_to_api(
            self._build_websocket_subscribe_request(access_token)
        )
        return await response.json()

    async def async_get_websocket_subscriptions(self, access_token: str) -> str:
        """Get websocket subscriptions."""
        response = await self._async_dict_to_api(
            self._build_websocket_get_request(access_token)
        )
        return await response.text()

    async def _async_dict_to_api(self, api_dict: dict[str, Any]) -> ClientResponse:
        url = api_dict.pop("url")
        method = api_dict.pop("method")
        access_token = api_dict.pop("access_token", None)
        payload = api_dict.get("params") or api_dict.get("json")

        if "headers" not in api_dict:
            api_dict["headers"] = _api_headers(
                access_token=access_token, brand=self.brand
            )

        if "version" in api_dict:
            api_dict["headers"][HEADER_ACCEPT_VERSION] = api_dict["version"]
            del api_dict["version"]

        if "timeout" not in api_dict:
            api_dict["timeout"] = self._timeout

        debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)

        if debug_enabled:
            _LOGGER.debug(
                "About to call %s with header=%s and payload=%s",
                url,
                _obscure_headers(api_dict["headers"]),
                _obscure_payload(payload),
            )

        attempts = 0
        while attempts < API_RETRY_ATTEMPTS:
            attempts += 1
            try:
                response = await self._aiohttp_session.request(method, url, **api_dict)
            except (
                ClientOSError,
                ClientSSLError,
                ServerDisconnectedError,
                ClientConnectionError,
            ) as ex:
                # Try again if we get disconnected
                # We may get [Errno 104] Connection reset by peer or a
                # transient disconnect/SSL error
                if attempts == API_RETRY_ATTEMPTS:
                    raise YaleApiError(
                        f"Failed to connect to August API: {ex}", ex
                    ) from ex
                await asyncio.sleep(API_EXCEPTION_RETRY_TIME)
                continue
            if debug_enabled:
                _LOGGER.debug(
                    "Received API response from url: %s, code: %s, headers: %s, content: %s",
                    url,
                    response.status,
                    _obscure_headers(response.headers),
                    await response.read(),
                )
            if response.status in (429, 502):
                # 429 - rate limited
                # 502 - bad gateway
                _LOGGER.debug(
                    "API sent a %s (attempt: %d), sleeping and trying again",
                    response.status,
                    attempts,
                )
                await asyncio.sleep(API_RETRY_TIME)
                continue
            break

        _raise_response_exceptions(response)

        return response


def _raise_response_exceptions(response: ClientResponse) -> None:
    """Raise exceptions for known error codes."""
    try:
        response.raise_for_status()
    except ClientResponseError as err:
        if err.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            raise InvalidAuth(
                f"Authentication failed: Verify brand is correct: {err.message}", err
            ) from err
        if err.status == 422:
            raise YaleApiError(
                f"The operation failed because the bridge (connect) is offline: {err.message}",
                err,
            ) from err
        if err.status == 423:
            raise YaleApiError(
                f"The operation failed because the bridge (connect) is in use: {err.message}",
                err,
            ) from err
        if err.status == 408:
            raise YaleApiError(
                f"The operation timed out because the bridge (connect) failed to respond: {err.message}",
                err,
            ) from err
        raise YaleApiError(
            f"The operation failed with error code {err.status}: {err.message}.", err
        ) from err



"""Connect to pubnub."""

import asyncio
import datetime
import logging
from collections.abc import Callable, Coroutine
from functools import partial
from typing import Any

from pubnub.callbacks import SubscribeCallback
from pubnub.enums import PNReconnectionPolicy, PNStatusCategory
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pubsub import PNMessageResult
from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub_asyncio import AsyncioSubscriptionManager, PubNubAsyncio

from .const import BRAND_CONFIG, Brand
from .device import DeviceDetail

_LOGGER = logging.getLogger(__name__)

UpdateCallbackType = Callable[[str, datetime.datetime, dict[str, Any]], None]

SHOULD_RECONNECT_CATEGORIES = {
    PNStatusCategory.PNUnknownCategory,
    PNStatusCategory.PNUnexpectedDisconnectCategory,
    PNStatusCategory.PNNetworkIssuesCategory,
    PNStatusCategory.PNTimeoutCategory,
}


class AugustPubNub(SubscribeCallback):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the AugustPubNub."""
        super().__init__(*args, **kwargs)
        self.connected = False
        self._device_channels: dict[str, str] = {}
        self._subscriptions: set[UpdateCallbackType] = set()

    def presence(self, pubnub: AsyncioSubscriptionManager, presence):
        _LOGGER.debug("Received new presence: %s", presence)

    def status(self, pubnub: AsyncioSubscriptionManager, status: PNStatus) -> None:
        if not pubnub:
            self.connected = False
            return

        _LOGGER.debug(
            "Received new status: category=%s error_data=%s error=%s status_code=%s operation=%s",
            status.category,
            status.error_data,
            status.error,
            status.status_code,
            status.operation,
        )

        if status.category in SHOULD_RECONNECT_CATEGORIES:
            self.connected = False
            pubnub.reconnect()

        elif status.category == PNStatusCategory.PNReconnectedCategory:
            self.connected = True
            now = datetime.datetime.now(datetime.UTC)
            # Callback with an empty message to force a refresh
            for callback in self._subscriptions:
                for device_id in self._device_channels.values():
                    callback(device_id, now, {})

        elif status.category == PNStatusCategory.PNConnectedCategory:
            self.connected = True

    def message(
        self, pubnub: AsyncioSubscriptionManager, message: PNMessageResult
    ) -> None:
        # Handle new messages
        device_id = self._device_channels[message.channel]
        _LOGGER.debug(
            "Received new messages on channel %s for device_id: %s with timetoken: %s: %s",
            message.channel,
            device_id,
            message.timetoken,
            message.message,
        )
        dt = datetime.datetime.fromtimestamp(
            int(message.timetoken) / 10000000, tz=datetime.timezone.utc
        )
        for callback in self._subscriptions:
            callback(device_id, dt, message.message)

    def subscribe(self, update_callback: UpdateCallbackType) -> Callable[[], None]:
        """Add an callback subscriber.

        Returns a callable that can be used to unsubscribe.
        """
        self._subscriptions.add(update_callback)
        return partial(self._unsubscribe, update_callback)

    def _unsubscribe(self, update_callback: UpdateCallbackType) -> None:
        self._subscriptions.remove(update_callback)

    def register_device(self, device_detail: DeviceDetail) -> None:
        """Register a device to get updates."""
        if device_detail.pubsub_channel is None:
            return
        self._device_channels[device_detail.pubsub_channel] = device_detail.device_id

    @property
    def channels(self):
        """Return a list of registered channels."""
        return self._device_channels.keys()

    async def run(
        self, user_uuid: str, brand: Brand = Brand.AUGUST
    ) -> Callable[[], Coroutine[Any, Any, None]]:
        """Run the pubnub loop."""
        brand_config = BRAND_CONFIG[brand]
        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = brand_config.pubnub_subscribe_token
        pnconfig.publish_key = brand_config.pubnub_publish_token
        pnconfig.uuid = f"pn-{str(user_uuid).upper()}"
        pnconfig.reconnect_policy = PNReconnectionPolicy.EXPONENTIAL
        pubnub = PubNubAsyncio(pnconfig)
        pubnub.add_listener(self)
        pubnub.subscribe().channels(self.channels).execute()

        async def _async_unsub():
            _LOGGER.debug("Removing listeners PubNub")
            pubnub.remove_listener(self)
            _LOGGER.debug("Unsubscribing from PubNub")
            pubnub.unsubscribe_all()
            await asyncio.sleep(0.1)  # Allow the unsubscribe to complete
            _LOGGER.debug("Stopping PubNub")
            await pubnub.stop()
            _LOGGER.debug("PubNub stopped")

        return _async_unsub



from __future__ import annotations

from http import HTTPStatus

from aiohttp import ClientError, ClientResponseError


class AugustApiAIOHTTPError(Exception):
    """An yale access api error with a friendly user consumable string."""

    def __init__(
        self,
        message: str | None = None,
        aiohttp_client_error: ClientError | None = None,
    ) -> None:
        """Initialize the error."""
        super().__init__(message or type(self).__name__)
        self.aiohttp_client_error = aiohttp_client_error
        self.status = (
            isinstance(aiohttp_client_error, ClientResponseError)
            and aiohttp_client_error.status
        )
        self.auth_failed = self.status in (
            HTTPStatus.UNAUTHORIZED,
            HTTPStatus.FORBIDDEN,
        )


class YaleApiError(AugustApiAIOHTTPError):
    """An yale access api error with a friendly user consumable string."""


class RequireValidation(Exception):
    """Error to indicate we require validation (2fa)."""


class CannotConnect(YaleApiError):
    """Error to indicate we cannot connect."""


class InvalidAuth(YaleApiError):
    """Error to indicate there is invalid auth."""


class RateLimited(YaleApiError):
    """Error to indicate we are rate limited."""

    def __init__(self, message: str, next_allowed: float) -> None:
        """Initialize the error."""
        super().__init__(message)
        self.next_allowed = next_allowed


class YaleXSError(Exception):
    """Base error."""


class ContentTokenExpired(Exception):
    """Token required for accessing this resource is not valid."""



import datetime
import json
import os
import unittest

import dateutil.parser

from yalexs.activity import (
    SOURCE_LOG,
    SOURCE_PUBNUB,
    BridgeOperationActivity,
    DoorbellMotionActivity,
    DoorOperationActivity,
    LockOperationActivity,
)
from yalexs.api_common import _convert_lock_result_to_activities
from yalexs.const import Brand
from yalexs.doorbell import DoorbellDetail
from yalexs.lock import LockDetail, LockDoorStatus, LockStatus
from yalexs.pubnub_activity import activities_from_pubnub_message
from yalexs.util import (
    as_utc_from_local,
    get_configuration_url,
    get_latest_activity,
    get_ssl_context,
    update_doorbell_image_from_activity,
    update_lock_detail_from_activity,
)


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path) as fptr:
        return fptr.read()


def test_get_latest_activity():
    """Test when two activities happen at the same time we prefer the one that is not moving."""
    lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
    unlocking_activities = activities_from_pubnub_message(
        lock,
        dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
        {
            "remoteEvent": 1,
            "status": "kAugLockState_Unlocking",
            "info": {
                "action": "unlock",
                "startTime": "2021-03-20T18:19:05.373Z",
                "context": {
                    "transactionID": "_oJRZKJsx",
                    "startDate": "2021-03-20T18:19:05.371Z",
                    "retryCount": 1,
                },
                "lockType": "lock_version_1001",
                "serialNumber": "M1FBA029QJ",
                "rssi": -53,
                "wlanRSSI": -55,
                "wlanSNR": 44,
                "duration": 2534,
            },
        },
    )
    unlocked_activities = activities_from_pubnub_message(
        lock,
        dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
        {
            "remoteEvent": 1,
            "status": "kAugLockState_Unlocked",
            "info": {
                "action": "unlock",
                "startTime": "2021-03-20T18:19:05.373Z",
                "context": {
                    "transactionID": "_oJRZKJsx",
                    "startDate": "2021-03-20T18:19:05.371Z",
                    "retryCount": 1,
                },
                "lockType": "lock_version_1001",
                "serialNumber": "M1FBA029QJ",
                "rssi": -53,
                "wlanRSSI": -55,
                "wlanSNR": 44,
                "duration": 2534,
            },
        },
    )
    assert get_latest_activity(unlocking_activities[0], None) == unlocking_activities[0]
    assert get_latest_activity(None, unlocked_activities[0]) == unlocked_activities[0]
    assert (
        get_latest_activity(unlocking_activities[0], unlocked_activities[0])
        == unlocked_activities[0]
    )
    assert (
        get_latest_activity(unlocked_activities[0], unlocking_activities[0])
        == unlocked_activities[0]
    )


def test_update_lock_detail_from_activity():
    """Test when two activities happen at the same time we prefer the one that is not moving."""
    lock = LockDetail(json.loads(load_fixture("get_lock.doorsense_init.json")))
    unlocking_activities = activities_from_pubnub_message(
        lock,
        dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
        {
            "remoteEvent": 1,
            "status": "kAugLockState_Unlocking",
            "info": {
                "action": "unlock",
                "startTime": "2021-03-20T18:19:05.373Z",
                "context": {
                    "transactionID": "_oJRZKJsx",
                    "startDate": "2021-03-20T18:19:05.371Z",
                    "retryCount": 1,
                },
                "lockType": "lock_version_1001",
                "serialNumber": "M1FBA029QJ",
                "rssi": -53,
                "wlanRSSI": -55,
                "wlanSNR": 44,
                "duration": 2534,
            },
        },
    )
    unlocked_activities = activities_from_pubnub_message(
        lock,
        dateutil.parser.parse("2017-12-10T05:48:30.272Z"),
        {
            "remoteEvent": 1,
            "status": "kAugLockState_Unlocked",
            "info": {
                "action": "unlock",
                "startTime": "2021-03-20T18:19:05.373Z",
                "context": {
                    "transactionID": "_oJRZKJsx",
                    "startDate": "2021-03-20T18:19:05.371Z",
                    "retryCount": 1,
                },
                "lockType": "lock_version_1001",
                "serialNumber": "M1FBA029QJ",
                "rssi": -53,
                "wlanRSSI": -55,
                "wlanSNR": 44,
                "duration": 2534,
            },
        },
    )
    update_lock_detail_from_activity(lock, unlocking_activities[0])
    assert lock.lock_status == LockStatus.UNLOCKING
    update_lock_detail_from_activity(lock, unlocked_activities[0])
    assert lock.lock_status == LockStatus.UNLOCKED
    update_lock_detail_from_activity(lock, unlocking_activities[0])
    assert lock.lock_status == LockStatus.UNLOCKED


class TestLockDetail(unittest.TestCase):
    def test_update_lock_with_activity_has_no_status(self):
        lock = LockDetail(
            json.loads(load_fixture("get_lock.nostatus_with_doorsense.json"))
        )
        self.assertEqual("ABC", lock.device_id)
        self.assertEqual(LockStatus.UNKNOWN, lock.lock_status)
        self.assertEqual(LockDoorStatus.UNKNOWN, lock.door_state)
        self.assertEqual(None, lock.lock_status_datetime)
        self.assertEqual(None, lock.door_state_datetime)

        unlock_operation_activity = LockOperationActivity(
            SOURCE_LOG, json.loads(load_fixture("unlock_activity.json"))
        )

        self.assertTrue(
            update_lock_detail_from_activity(lock, unlock_operation_activity)
        )
        self.assertEqual(LockStatus.UNLOCKED, lock.lock_status)

    def test_update_lock_with_activity(self):
        lock = LockDetail(
            json.loads(load_fixture("get_lock.online_with_doorsense.json"))
        )
        self.assertEqual("ABC", lock.device_id)
        self.assertEqual(LockStatus.LOCKED, lock.lock_status)
        self.assertEqual(LockDoorStatus.OPEN, lock.door_state)
        self.assertEqual(
            dateutil.parser.parse("2017-12-10T04:48:30.272Z"), lock.lock_status_datetime
        )
        self.assertEqual(
            dateutil.parser.parse("2017-12-10T04:48:30.272Z"), lock.door_state_datetime
        )

        lock_operation_activity = LockOperationActivity(
            SOURCE_LOG, json.loads(load_fixture("lock_activity.json"))
        )
        unlock_operation_activity = LockOperationActivity(
            SOURCE_LOG, json.loads(load_fixture("unlock_activity.json"))
        )
        open_operation_activity = DoorOperationActivity(
            SOURCE_LOG, json.loads(load_fixture("door_open_activity.json"))
        )
        closed_operation_activity = DoorOperationActivity(
            SOURCE_LOG, json.loads(load_fixture("door_closed_activity.json"))
        )
        closed_operation_wrong_deviceid_activity = DoorOperationActivity(
            SOURCE_LOG,
            json.loads(load_fixture("door_closed_activity_wrong_deviceid.json")),
        )
        closed_operation_wrong_houseid_activity = DoorOperationActivity(
            SOURCE_LOG,
            json.loads(load_fixture("door_closed_activity_wrong_houseid.json")),
        )

        self.assertTrue(
            update_lock_detail_from_activity(lock, unlock_operation_activity)
        )
        self.assertEqual(LockStatus.UNLOCKED, lock.lock_status)
        self.assertEqual(
            as_utc_from_local(datetime.datetime.fromtimestamp(1582007217000 / 1000)),
            lock.lock_status_datetime,
        )

        self.assertTrue(update_lock_detail_from_activity(lock, lock_operation_activity))
        self.assertEqual(LockStatus.LOCKED, lock.lock_status)
        self.assertEqual(
            as_utc_from_local(datetime.datetime.fromtimestamp(1582007218000 / 1000)),
            lock.lock_status_datetime,
        )

        # returns false we send an older activity
        self.assertFalse(
            update_lock_detail_from_activity(lock, unlock_operation_activity)
        )

        self.assertTrue(
            update_lock_detail_from_activity(lock, closed_operation_activity)
        )
        self.assertEqual(LockDoorStatus.CLOSED, lock.door_state)
        self.assertEqual(
            as_utc_from_local(datetime.datetime.fromtimestamp(1582007217000 / 1000)),
            lock.door_state_datetime,
        )

        self.assertTrue(update_lock_detail_from_activity(lock, open_operation_activity))
        self.assertEqual(LockDoorStatus.OPEN, lock.door_state)
        self.assertEqual(
            as_utc_from_local(datetime.datetime.fromtimestamp(1582007219000 / 1000)),
            lock.door_state_datetime,
        )

        # returns false we send an older activity
        self.assertFalse(
            update_lock_detail_from_activity(lock, closed_operation_activity)
        )

        with self.assertRaises(ValueError):
            update_lock_detail_from_activity(
                lock, closed_operation_wrong_deviceid_activity
            )

        # We do not always have the houseid so we do not throw
        # as long as the deviceid is correct since they are unique
        self.assertFalse(
            update_lock_detail_from_activity(
                lock, closed_operation_wrong_houseid_activity
            )
        )

        self.assertEqual(LockDoorStatus.OPEN, lock.door_state)
        self.assertEqual(LockStatus.LOCKED, lock.lock_status)
        activities = _convert_lock_result_to_activities(
            json.loads(load_fixture("unlock.json"))
        )
        for activity in activities:
            self.assertTrue(update_lock_detail_from_activity(lock, activity))
        self.assertEqual(LockDoorStatus.CLOSED, lock.door_state)
        self.assertEqual(LockStatus.UNLOCKED, lock.lock_status)

        bridge_offline_activity = BridgeOperationActivity(
            SOURCE_PUBNUB,
            {
                "action": "associated_bridge_offline",
                "callingUser": {"UserID": None},
                "dateTime": 1512906510272.0,
                "deviceName": "Front Door Lock",
                "deviceType": "lock",
                "deviceID": lock.device_id,
                "house": "000000000000",
                "info": {},
            },
        )
        assert bridge_offline_activity.source == SOURCE_PUBNUB
        self.assertTrue(update_lock_detail_from_activity(lock, bridge_offline_activity))
        assert lock.bridge_is_online is False
        bridge_online_activity = BridgeOperationActivity(
            SOURCE_PUBNUB,
            {
                "action": "associated_bridge_online",
                "callingUser": {"UserID": None},
                "dateTime": 1512906510272.0,
                "deviceName": "Front Door Lock",
                "deviceType": "lock",
                "deviceID": lock.device_id,
                "house": "000000000000",
                "info": {},
            },
        )
        self.assertTrue(update_lock_detail_from_activity(lock, bridge_online_activity))
        assert lock.bridge_is_online is True
        assert bridge_online_activity.source == SOURCE_PUBNUB


class TestDetail(unittest.TestCase):
    def test_update_doorbell_image_from_activity(self):
        doorbell = DoorbellDetail(json.loads(load_fixture("get_doorbell.json")))
        self.assertEqual("K98GiDT45GUL", doorbell.device_id)
        self.assertEqual(
            dateutil.parser.parse("2017-12-10T08:01:35Z"),
            doorbell.image_created_at_datetime,
        )
        self.assertEqual(
            "https://image.com/vmk16naaaa7ibuey7sar.jpg", doorbell.image_url
        )
        doorbell_motion_activity_no_image = DoorbellMotionActivity(
            SOURCE_LOG,
            json.loads(load_fixture("doorbell_motion_activity_no_image.json")),
        )
        self.assertFalse(
            update_doorbell_image_from_activity(
                doorbell, doorbell_motion_activity_no_image
            )
        )
        doorbell_motion_activity = DoorbellMotionActivity(
            SOURCE_LOG, json.loads(load_fixture("doorbell_motion_activity.json"))
        )
        self.assertTrue(
            update_doorbell_image_from_activity(doorbell, doorbell_motion_activity)
        )
        self.assertEqual(
            dateutil.parser.parse("2020-02-20T17:44:45Z"),
            doorbell.image_created_at_datetime,
        )
        self.assertEqual("https://my.updated.image/image.jpg", doorbell.image_url)
        old_doorbell_motion_activity = DoorbellMotionActivity(
            SOURCE_LOG, json.loads(load_fixture("doorbell_motion_activity_old.json"))
        )
        # returns false we send an older activity
        self.assertFalse(
            update_doorbell_image_from_activity(doorbell, old_doorbell_motion_activity)
        )
        self.assertEqual(
            dateutil.parser.parse("2020-02-20T17:44:45Z"),
            doorbell.image_created_at_datetime,
        )
        self.assertEqual("https://my.updated.image/image.jpg", doorbell.image_url)
        wrong_doorbell_motion_activity = DoorbellMotionActivity(
            SOURCE_LOG, json.loads(load_fixture("doorbell_motion_activity_wrong.json"))
        )

        with self.assertRaises(ValueError):
            update_doorbell_image_from_activity(
                doorbell, wrong_doorbell_motion_activity
            )

    def test_update_doorbell_image_from_activity_missing_image_at_start(self):
        doorbell = DoorbellDetail(
            json.loads(load_fixture("get_doorbell_missing_image.json"))
        )
        self.assertEqual("K98GiDT45GUL", doorbell.device_id)
        self.assertEqual(
            None,
            doorbell.image_created_at_datetime,
        )
        self.assertEqual(None, doorbell.image_url)
        doorbell_motion_activity_no_image = DoorbellMotionActivity(
            SOURCE_LOG,
            json.loads(load_fixture("doorbell_motion_activity_no_image.json")),
        )
        self.assertFalse(
            update_doorbell_image_from_activity(
                doorbell, doorbell_motion_activity_no_image
            )
        )
        doorbell_motion_activity = DoorbellMotionActivity(
            SOURCE_LOG, json.loads(load_fixture("doorbell_motion_activity.json"))
        )
        self.assertTrue(
            update_doorbell_image_from_activity(doorbell, doorbell_motion_activity)
        )
        self.assertEqual(
            dateutil.parser.parse("2020-02-20T17:44:45Z"),
            doorbell.image_created_at_datetime,
        )
        self.assertEqual("https://my.updated.image/image.jpg", doorbell.image_url)


def test_get_configuration_url():
    """Test that we get the correct configuration url for the brand."""
    assert get_configuration_url("august") == "https://account.august.com"
    assert get_configuration_url("yale_access") == "https://account.august.com"
    assert get_configuration_url("yale_home") == "https://account.aaecosystem.com"
    assert get_configuration_url(Brand.AUGUST) == "https://account.august.com"
    assert get_configuration_url(Brand.YALE_ACCESS) == "https://account.august.com"
    assert get_configuration_url(Brand.YALE_HOME) == "https://account.aaecosystem.com"


def test_get_ssl_context():
    """Test getting the ssl context is cached."""
    assert get_ssl_context() is get_ssl_context()
